use std::ffi::{c_void, CStr, CString};
use std::os::raw::c_char;

#[repr(C)]
struct tof_error {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_error_delete(_: *mut tof_error);
    fn tof_error_message(_: *mut tof_error) -> *const c_char;
}

pub struct ErrorHelper {
    ptr: *mut tof_error,
}

impl Drop for ErrorHelper {
    fn drop(&mut self) {
        unsafe { tof_error_delete(self.ptr) }
    }
}

impl ErrorHelper {
    fn new() -> Self {
        Self {
            ptr: std::ptr::null_mut(),
        }
    }

    fn ptr(&mut self) -> *mut *mut tof_error {
        std::ptr::addr_of_mut!(self.ptr)
    }

    fn has_ptr(&self) -> bool {
        !self.ptr.is_null()
    }

    fn make_error(&self) -> Box<dyn std::error::Error> {
        let msg = unsafe {
            let ret = tof_error_message(self.ptr);
            CStr::from_ptr(ret)
        };

        let str = match msg.to_str() {
            Ok(ok) => ok,
            Err(err) => return Box::new(err),
        };

        Error::new(&str.to_owned())
    }

    fn make_result<T>(self, input: T) -> Result<T> {
        if self.has_ptr() {
            Err(self.make_error())
        } else {
            Ok(input)
        }
    }

    fn empty_result(self) -> Result<()> {
        self.make_result(())
    }
}

#[derive(Debug)]
pub struct Error {
    pub message: String,
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", self.message)
    }
}

impl Error {
    fn new(str: &str) -> Box<dyn std::error::Error> {
        Box::new(Error {
            message: str.to_owned(),
        })
    }
}

impl std::error::Error for Error {}

pub type Result<T> = std::result::Result<T, Box<dyn std::error::Error>>;

mod private {
    pub trait Private {}
}

/// The different frame types that are supported
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum FrameType {
    /// The 16bits coming off the OPT8241 image sensor
    RawCommon = 0,
    /// Common (A + B) image, 4 or 12 bits depending on the sensor
    Common = 1,
    /// Raw (A - B) image 12 bits signed data
    Raw = 2,
    /// The real image
    Real = 3,
    /// The imaginary image
    Imaginary = 4,
    /// The phase image
    Phase = 5,
    /// The amplitude image
    Amplitude = 6,
    /// The radial distance image
    Radial = 7,
    /// The intensity image
    Intensity = 8,
    /// The x image of the point cloud
    X = 9,
    /// The y image of the point cloud
    Y = 10,
    /// The z image of the point cloud
    Z = 11,
    /// A complete pointcloud image: float x, float y, float z, float padding
    Xyz = 12,
    /// A complete pointcloud image with amplitude: float x, float y, float z,
    /// float amplitude
    XyzAmp = 13,
    /// A complete pointcloud image with bgr: float x, float y, float z, uint8
    /// b, uint8 g, uint8 r, uint8 padding
    XyzBgr = 14,
    /// A complete pointcloud image with bgr and intensity: float x, float y,
    /// float z, uint8 b, uint8 g, uint8 r, uint8 intensity
    XyzBgrI = 15,
    /// A color image, encode bgr
    Bgr = 18,
    /// A color image encoded in YUV422
    Yuv = 19,
    /// A raw image containing both the A and B tap
    RawAb = 20,
    /// a color image encoded in mjpeg
    Mjpeg = 21,
    /// A color image projected to the XYZ locations of the depth sensor
    BgrProjected = 22,
    /// The averaged value for each dot. In the XYZAmp format.
    CompressedDot = 23,
}

#[repr(C)]
struct tof_stream {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_stream_delete(_: *mut tof_stream);
    fn tof_stream_frame_type(_: *const tof_stream, _: *mut *mut tof_error) -> i32;
    fn tof_stream_frame_id(_: *const tof_stream, _: *mut *mut tof_error) -> u32;
    fn tof_stream_modulation_frequency(_: *const tof_stream, _: *mut *mut tof_error) -> f32;
    fn tof_stream_integration_time(_: *const tof_stream, _: *mut *mut tof_error) -> u32;
    fn tof_stream_process(_: *const tof_stream, _: *mut *mut tof_error) -> u32;
}

/// This class describes the different data streams coming from the camera
pub struct Stream {
    ptr: *mut tof_stream,
}

impl Drop for Stream {
    fn drop(&mut self) {
        unsafe { tof_stream_delete(self.ptr); }
    }
}

unsafe impl Send for Stream {}

impl Stream {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_stream {
        self.ptr
    }

    /// Get the frame type
    ///
    /// ### Returns
    ///
    /// The frame type
    ///
    pub fn frame_type(&self) -> Result<FrameType> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_stream_frame_type(self.ptr, err.ptr()) };
        let out: FrameType = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Get the frame id
    ///
    /// ### Returns
    ///
    /// The frame id
    ///
    pub fn frame_id(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_stream_frame_id(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the modulation frequency of the frame
    ///
    /// ### Returns
    ///
    /// The modulation frequency
    ///
    pub fn modulation_frequency(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_stream_modulation_frequency(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the integration time of the frame
    ///
    /// ### Returns
    ///
    /// The integration time
    ///
    pub fn integration_time(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_stream_integration_time(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get a bit array of the processing performed on the frame
    ///
    /// ### Returns
    ///
    /// A bit array
    ///
    pub fn process(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_stream_process(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_frame_type_to_string(_: i32, _: *mut *mut tof_error) -> *const c_char;
}

/// Get the string representation of the frame type
///
/// ### Arguments
///
/// - `frame_type` Frame type
///
/// ### Returns
///
/// String representation of frame type
///
pub fn frame_type_to_string(frame_type: FrameType) -> Result<&'static str> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_frame_type_to_string(frame_type as i32, err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
    Ok(unsafe { CStr::from_ptr(ret).to_str()? })
    }
}

pub type UserPointerDestructed = extern "C" fn(*mut u8, usize, *mut c_void);

/// The different matrix types
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum MatType {
    /// A single unsigned byte
    Uint8 = 0,
    /// A single signed byte
    Int8 = 1,
    /// Two unsigned bytes
    Uint16 = 2,
    /// Two signed bytes
    Int16 = 3,
    /// Four signed bytes
    Int32 = 4,
    /// Four bytes floating point value
    Float = 5,
    /// Eight bytes floating point value
    Double = 6,
    /// Four bytes floating point value with four channels
    Float4 = 29,
    /// Four bytes floating point value with eight channels
    Float8 = 61,
}

#[repr(C)]
struct tof_data {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_data_delete(_: *mut tof_data);
    fn tof_data_data(_: *const tof_data, _: *mut *mut tof_error) -> *mut c_void;
    fn tof_data_size(_: *const tof_data, _: *mut *mut tof_error) -> usize;
    fn tof_data_get_user_data(_: *const tof_data, _: *mut *mut tof_error) -> u64;
    fn tof_data_set_user_data(_: *mut tof_data, _: u64, _: *mut *mut tof_error);
    fn tof_data_mat_type(_: *const tof_data, _: *mut *mut tof_error) -> i32;
    fn tof_data_rows(_: *const tof_data, _: *mut *mut tof_error) -> i32;
    fn tof_data_cols(_: *const tof_data, _: *mut *mut tof_error) -> i32;
    fn tof_data_frame_count(_: *const tof_data, _: *mut *mut tof_error) -> u16;
    fn tof_data_frame_type(_: *const tof_data, _: *mut *mut tof_error) -> i32;
    fn tof_data_frame_id(_: *const tof_data, _: *mut *mut tof_error) -> u32;
    fn tof_data_modulation_frequency(_: *const tof_data, _: *mut *mut tof_error) -> f32;
    fn tof_data_integration_time(_: *const tof_data, _: *mut *mut tof_error) -> u32;
    fn tof_data_process(_: *const tof_data, _: *mut *mut tof_error) -> u32;
    fn tof_data_scale(_: *const tof_data, _: *mut *mut tof_error) -> f32;
    fn tof_data_addition(_: *const tof_data, _: *mut *mut tof_error) -> f32;
    fn tof_data_temperatures(_: *const tof_data, _: *const f32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_data_sync_count(_: *const tof_data, _: *mut *mut tof_error) -> u8;
    fn tof_data_user_allocated(_: *const tof_data, _: *mut *mut tof_error) -> bool;
    fn tof_data_user_pointer(_: *const tof_data, _: *mut *mut tof_error) -> *mut c_void;
    fn tof_data_acceleration(_: *const tof_data, _: *mut *mut tof_error) -> *const i16;
    fn tof_data_angular_velocity(_: *const tof_data, _: *mut *mut tof_error) -> *const i16;
}

/// This is the class that contains depth or image data
pub struct Data {
    ptr: *mut tof_data,
}

impl Drop for Data {
    fn drop(&mut self) {
        unsafe { tof_data_delete(self.ptr); }
    }
}

unsafe impl Send for Data {}

impl Data {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_data {
        self.ptr
    }

    /// Get pointer to data
    ///
    /// ### Returns
    ///
    /// 
    ///
    pub fn data(&self) -> Result<*mut c_void> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get size of the data
    ///
    /// ### Returns
    ///
    /// The size
    ///
    pub fn size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get user data that will get saved to csf file
    ///
    /// ### Returns
    ///
    /// Set user data
    ///
    pub fn get_user_data(&self) -> Result<u64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_get_user_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set user data that will get saved to csf file
    ///
    /// ### Arguments
    ///
    /// - `user_data` Set user data
    ///
    pub fn set_user_data(&mut self, user_data: u64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_data_set_user_data(self.ptr, user_data, err.ptr()) };
        err.empty_result()
    }

    /// Get the matrix type
    ///
    /// ### Returns
    ///
    /// The matrix type
    ///
    pub fn mat_type(&self) -> Result<MatType> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_mat_type(self.ptr, err.ptr()) };
        let out: MatType = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Get the number of rows
    ///
    /// ### Returns
    ///
    /// Number of rows
    ///
    pub fn rows(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_rows(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the number of columns
    ///
    /// ### Returns
    ///
    /// Number of columns
    ///
    pub fn cols(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_cols(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// The frame counter
    ///
    /// ### Returns
    ///
    /// Frame count
    ///
    pub fn frame_count(&self) -> Result<u16> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_frame_count(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the frame type
    ///
    /// ### Returns
    ///
    /// The frame type
    ///
    pub fn frame_type(&self) -> Result<FrameType> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_frame_type(self.ptr, err.ptr()) };
        let out: FrameType = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Get the frame id
    ///
    /// ### Returns
    ///
    /// The frame id
    ///
    pub fn frame_id(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_frame_id(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the modulation frequency of the frame
    ///
    /// ### Returns
    ///
    /// The modulation frequency
    ///
    pub fn modulation_frequency(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_modulation_frequency(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the integration time of the frame
    ///
    /// ### Returns
    ///
    /// The integration time
    ///
    pub fn integration_time(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_integration_time(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get a bit array of the processing performed on the frame
    ///
    /// ### Returns
    ///
    /// A bit array
    ///
    pub fn process(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_process(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the scale used for integer frames, to get the original value divide
    /// by scale
    ///
    /// ### Returns
    ///
    /// The scale
    ///
    pub fn scale(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_scale(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the addition used for integer frames, to get the original value
    /// substract the addition
    ///
    /// ### Returns
    ///
    /// The addition
    ///
    pub fn addition(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_addition(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the temperature of various parts of the camera
    ///
    /// ### Returns
    ///
    /// Temperatures
    ///
    pub fn temperatures(&self) -> Result<Vec<f32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_data_temperatures(self.ptr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<f32> = Vec::with_capacity(size);
        let ret = unsafe { tof_data_temperatures(self.ptr, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the synchronization count
    ///
    /// ### Returns
    ///
    /// Sync count
    ///
    pub fn sync_count(&self) -> Result<u8> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_sync_count(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Check whether the data uses a pointer supplied by the user
    ///
    /// ### Returns
    ///
    /// Data allocated by user
    ///
    pub fn user_allocated(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_user_allocated(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// The user data that was given in the add_user_pointer function
    ///
    /// ### Returns
    ///
    /// User pointer
    ///
    pub fn user_pointer(&self) -> Result<*mut c_void> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_user_pointer(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// The acceleration measured by the imu
    ///
    /// ### Returns
    ///
    /// Acceleration, x y z
    ///
    pub fn acceleration(&self) -> Result<[i16; 3]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_acceleration(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[i16; 3]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 3);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// The angular velocity measured by the imu
    ///
    /// ### Returns
    ///
    /// Angular velocity, x y z
    ///
    pub fn angular_velocity(&self) -> Result<[i16; 3]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_data_angular_velocity(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[i16; 3]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 3);
            arr.assume_init()
        };
        err.make_result(arr)
    }

}

#[repr(C)]
struct tof_csf_reader {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_csf_reader_delete(_: *mut tof_csf_reader);
    fn tof_csf_reader_new(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_csf_reader;
    fn tof_csf_reader_read_frame(_: *mut tof_csf_reader, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_csf_reader_read_frame_into_pointer(_: *mut tof_csf_reader, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_csf_reader_number_of_frames(_: *const tof_csf_reader, _: *mut *mut tof_error) -> usize;
    fn tof_csf_reader_get_header(_: *const tof_csf_reader, _: usize, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_csf_reader_next_stream(_: *const tof_csf_reader, _: *mut *mut tof_error) -> *mut tof_stream;
    fn tof_csf_reader_next_frame_size(_: *const tof_csf_reader, _: *mut *mut tof_error) -> usize;
    fn tof_csf_reader_frame_index(_: *const tof_csf_reader, _: *mut *mut tof_error) -> usize;
    fn tof_csf_reader_go_to_frame(_: *mut tof_csf_reader, _: usize, _: *mut *mut tof_error);
    fn tof_csf_reader_streams(_: *const tof_csf_reader, _: *const *mut tof_stream, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_csf_reader_stream_count(_: *const tof_csf_reader, _: *mut tof_stream, _: *mut *mut tof_error) -> usize;
    fn tof_csf_reader_go_to_stream(_: *mut tof_csf_reader, _: *mut tof_stream, _: usize, _: *mut *mut tof_error);
    fn tof_csf_reader_get_calibration(_: *const tof_csf_reader, _: *mut *mut tof_error) -> *mut tof_calibration;
    fn tof_csf_reader_get_camera_config(_: *const tof_csf_reader, _: *mut *mut tof_error) -> *mut tof_camera_config;
}

/// DEPRECATED, use CSFCamera. This class reads csf files
pub struct CsfReader {
    ptr: *mut tof_csf_reader,
}

impl Drop for CsfReader {
    fn drop(&mut self) {
        unsafe { tof_csf_reader_delete(self.ptr); }
    }
}

unsafe impl Send for CsfReader {}

impl CsfReader {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_csf_reader {
        self.ptr
    }

    /// Open a csf file for reading
    ///
    /// ### Arguments
    ///
    /// - `file_location` File location of the csf file
    ///
    pub fn new(file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_csf_reader_new(file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CsfReader { ptr: ret })
        }
    }

    /// Read the next data frame from disk
    ///
    /// ### Returns
    ///
    /// A data frame
    ///
    pub fn read_frame(&mut self) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_read_frame(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Read the next data frame from disk into the supplied pointer
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    ///
    /// ### Returns
    ///
    /// A data frame
    ///
    pub unsafe fn read_frame_into_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = tof_csf_reader_read_frame_into_pointer(self.ptr, pointer, capacity, callback, callback_user_ptr, err.ptr());
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get the total number of frames in the csf file
    ///
    /// ### Returns
    ///
    /// Total number of frames in the csf file
    ///
    pub fn number_of_frames(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_number_of_frames(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Returns an empty Data class containing the header information for the
    /// specified frame
    ///
    /// ### Arguments
    ///
    /// - `index` Which frame to get the header from
    ///
    /// ### Returns
    ///
    /// Empty Data instance containing header information
    ///
    pub fn get_header(&self, index: usize) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_get_header(self.ptr, index, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get the stream information of the next frame
    ///
    /// ### Returns
    ///
    /// Information on the next frame
    ///
    pub fn next_stream(&self) -> Result<Stream> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_next_stream(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Stream { ptr: ret })
        }
    }

    /// Get the size of the next frame
    ///
    /// ### Returns
    ///
    /// Size of the next frame
    ///
    pub fn next_frame_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_next_frame_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the current frame position
    ///
    /// ### Returns
    ///
    /// Frame index
    ///
    pub fn frame_index(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_frame_index(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Go to specified frame
    ///
    /// ### Arguments
    ///
    /// - `index` The index of the frame to jump to
    ///
    pub fn go_to_frame(&mut self, index: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_csf_reader_go_to_frame(self.ptr, index, err.ptr()) };
        err.empty_result()
    }

    /// Get the streams that are in the csf file
    ///
    /// ### Returns
    ///
    /// Streams in the csf file
    ///
    pub fn streams(&self) -> Result<Vec<Stream>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_csf_reader_streams(self.ptr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<Stream> = Vec::with_capacity(size);
        let ret = unsafe { tof_csf_reader_streams(self.ptr, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the number of frames contained in the file of the given stream
    ///
    /// ### Arguments
    ///
    /// - `stream` The stream
    ///
    /// ### Returns
    ///
    /// Number of frames for given stream
    ///
    pub fn stream_count(&self, stream: &Stream) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_stream_count(self.ptr, stream.ptr(), err.ptr()) };
        err.make_result(ret)
    }

    /// Go to the specified stream, the next read frame call will start from
    /// here
    ///
    /// ### Arguments
    ///
    /// - `stream` The stream
    /// - `index` The frame index
    ///
    pub fn go_to_stream(&mut self, stream: &Stream, index: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_csf_reader_go_to_stream(self.ptr, stream.ptr(), index, err.ptr()) };
        err.empty_result()
    }

    /// Get calibration from the csf file
    ///
    /// ### Returns
    ///
    /// Calibration
    ///
    pub fn get_calibration(&self) -> Result<Calibration> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_get_calibration(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Calibration { ptr: ret })
        }
    }

    /// Get camera config from the csf file
    ///
    /// ### Returns
    ///
    /// Camera config
    ///
    pub fn get_camera_config(&self) -> Result<CameraConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_reader_get_camera_config(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

}

#[repr(C)]
struct tof_csf_writer {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_csf_writer_delete(_: *mut tof_csf_writer);
    fn tof_csf_writer_new(_: *const c_char, _: *const tof_camera_config, _: *const tof_calibration, _: *const c_char, _: *mut *mut tof_error) -> *mut tof_csf_writer;
    fn tof_csf_writer_new_from_camera(_: *const c_char, _: *mut tof_camera, _: *mut *mut tof_error) -> *mut tof_csf_writer;
    fn tof_csf_writer_write_frame(_: *mut tof_csf_writer, _: *const tof_data, _: *mut *mut tof_error);
}

/// The csf writer class takes care of writing depth frames to disk.  The class
/// has to be created from either a Camera or csf reader, because it needs
/// information about the camera that was used to record the depth frames. 
/// Therefore you should only write depth frames coming from one camera into the
/// same writer

pub struct CsfWriter {
    ptr: *mut tof_csf_writer,
}

impl Drop for CsfWriter {
    fn drop(&mut self) {
        unsafe { tof_csf_writer_delete(self.ptr); }
    }
}

unsafe impl Send for CsfWriter {}

impl CsfWriter {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_csf_writer {
        self.ptr
    }

    /// Construct csf writer from calibration and configuration
    ///
    /// ### Arguments
    ///
    /// - `file_location` Path to where the csf file should be created
    /// - `camera_config` Camera configuration
    /// - `calibration` Calibration
    /// - `serial` Serial number of camera
    ///
    pub fn new(file_location: &str, camera_config: &CameraConfig, calibration: &Calibration, serial: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let serial_str = CString::new(serial)?;
        let ret = unsafe { tof_csf_writer_new(file_location_str.as_ptr(), camera_config.ptr(), calibration.ptr(), serial_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CsfWriter { ptr: ret })
        }
    }

    /// Construct the csf writer from a camera
    ///
    /// ### Arguments
    ///
    /// - `file_location` Path to where the csf file should be created
    /// - `camera` The camera to get the camera config and calibration file from
    ///
    pub fn new_from_camera(file_location: &str, camera: &impl Camera) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_csf_writer_new_from_camera(file_location_str.as_ptr(), camera.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CsfWriter { ptr: ret })
        }
    }

    /// Write frame to csf file
    ///
    /// ### Arguments
    ///
    /// - `frame` A data frame
    ///
    pub fn write_frame(&mut self, frame: &Data) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_csf_writer_write_frame(self.ptr, frame.ptr(), err.ptr()) };
        err.empty_result()
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_create_csf_writer_reader(_: *const c_char, _: *const tof_csf_reader, _: *mut *mut tof_error) -> *mut tof_csf_writer;
    fn tof_create_csf_writer_camera(_: *const c_char, _: *const tof_camera, _: *mut *mut tof_error) -> *mut tof_csf_writer;
    fn tof_create_csf_writer_gige_interface(_: *const c_char, _: *const tof_gige_interface, _: *mut *mut tof_error) -> *mut tof_csf_writer;
    fn tof_create_csf_writer_usb_interface(_: *const c_char, _: *const tof_usb_interface, _: *mut *mut tof_error) -> *mut tof_csf_writer;
}

/// Create csf writer from csf reader
///
/// ### Arguments
///
/// - `file_location` Path to where the csf file should be created
/// - `reader` A csf reader
///
/// ### Returns
///
/// csf writer
///
pub fn create_csf_writer_reader(file_location: &str, reader: &CsfReader) -> Result<CsfWriter> {
    let mut err = ErrorHelper::new();
    let file_location_str = CString::new(file_location)?;
    let ret = unsafe { tof_create_csf_writer_reader(file_location_str.as_ptr(), reader.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(CsfWriter { ptr: ret })
    }
}

/// Create csf writer from camera
///
/// ### Arguments
///
/// - `file_location` Path to where the csf file should be created
/// - `camera` A camera
///
/// ### Returns
///
/// csf writer
///
pub fn create_csf_writer_camera(file_location: &str, camera: &impl Camera) -> Result<CsfWriter> {
    let mut err = ErrorHelper::new();
    let file_location_str = CString::new(file_location)?;
    let ret = unsafe { tof_create_csf_writer_camera(file_location_str.as_ptr(), camera.camera_ptr() as *mut tof_camera, err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(CsfWriter { ptr: ret })
    }
}

/// Create csf writer from the GigE interface
///
/// ### Arguments
///
/// - `file_location` Path to where the csf file should be created
/// - `gige_interface` A GigE interface
///
/// ### Returns
///
/// csf writer
///
pub fn create_csf_writer_gige_interface(file_location: &str, gige_interface: &GigeInterface) -> Result<CsfWriter> {
    let mut err = ErrorHelper::new();
    let file_location_str = CString::new(file_location)?;
    let ret = unsafe { tof_create_csf_writer_gige_interface(file_location_str.as_ptr(), gige_interface.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(CsfWriter { ptr: ret })
    }
}

/// Create csf writer from the USB interface
///
/// ### Arguments
///
/// - `file_location` Path to where the csf file should be created
/// - `usb_interface` A USB interface
///
/// ### Returns
///
/// csf writer
///
pub fn create_csf_writer_usb_interface(file_location: &str, usb_interface: &UsbInterface) -> Result<CsfWriter> {
    let mut err = ErrorHelper::new();
    let file_location_str = CString::new(file_location)?;
    let ret = unsafe { tof_create_csf_writer_usb_interface(file_location_str.as_ptr(), usb_interface.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(CsfWriter { ptr: ret })
    }
}

/// The integration time as fixed presets
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum IntegrationTime {
    /// 250 us
    Short = 0,
    /// 500 us
    Medium = 1,
    /// 1000us
    Long = 2,
}

/// The different imaging environments
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum ImagingEnvironment {
    /// Inside buildings
    Inside = 0,
    /// When bright sunlight is present
    Sunlight = 1,
}

/// The optimization of the camera configuration
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum Strategy {
    /// The balanced configuration
    Balanced = 0,
    /// Optimize for moving objects
    Speed = 1,
    /// Optimize for high accuracy
    Accuracy = 2,
}

#[repr(C)]
struct tof_user_config {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_user_config_delete(_: *mut tof_user_config);
    fn tof_user_config_new_default(_: *mut *mut tof_error) -> *mut tof_user_config;
    fn tof_user_config_to_camera_config(_: *const tof_user_config, _: *mut tof_camera, _: *mut *mut tof_error) -> *mut tof_camera_config;
    fn tof_user_config_get_environment(_: *const tof_user_config, _: *mut *mut tof_error) -> i32;
    fn tof_user_config_set_environment(_: *mut tof_user_config, _: i32, _: *mut *mut tof_error);
    fn tof_user_config_get_max_distance(_: *const tof_user_config, _: *mut *mut tof_error) -> f32;
    fn tof_user_config_set_max_distance(_: *mut tof_user_config, _: f32, _: *mut *mut tof_error);
    fn tof_user_config_get_fps(_: *const tof_user_config, _: *mut *mut tof_error) -> f32;
    fn tof_user_config_set_fps(_: *mut tof_user_config, _: f32, _: *mut *mut tof_error);
    fn tof_user_config_get_lens_flare(_: *const tof_user_config, _: *mut *mut tof_error) -> bool;
    fn tof_user_config_set_lens_flare(_: *mut tof_user_config, _: bool, _: *mut *mut tof_error);
    fn tof_user_config_get_translucent(_: *const tof_user_config, _: *mut *mut tof_error) -> bool;
    fn tof_user_config_set_translucent(_: *mut tof_user_config, _: bool, _: *mut *mut tof_error);
    fn tof_user_config_get_hdr(_: *const tof_user_config, _: *mut *mut tof_error) -> bool;
    fn tof_user_config_set_hdr(_: *mut tof_user_config, _: bool, _: *mut *mut tof_error);
    fn tof_user_config_get_integration_time(_: *const tof_user_config, _: *mut *mut tof_error) -> i32;
    fn tof_user_config_set_integration_time(_: *mut tof_user_config, _: i32, _: *mut *mut tof_error);
    fn tof_user_config_get_strategy(_: *const tof_user_config, _: *mut *mut tof_error) -> i32;
    fn tof_user_config_set_strategy(_: *mut tof_user_config, _: i32, _: *mut *mut tof_error);
    fn tof_user_config_get_channel(_: *const tof_user_config, _: *mut *mut tof_error) -> i32;
    fn tof_user_config_set_channel(_: *mut tof_user_config, _: i32, _: *mut *mut tof_error);
}

/// Simple interface for camera configuration generation
pub struct UserConfig {
    ptr: *mut tof_user_config,
}

impl Drop for UserConfig {
    fn drop(&mut self) {
        unsafe { tof_user_config_delete(self.ptr); }
    }
}

unsafe impl Send for UserConfig {}

impl UserConfig {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_user_config {
        self.ptr
    }

    /// Create default user config
    ///
    pub fn new_default() -> Result<Self> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_new_default(err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(UserConfig { ptr: ret })
        }
    }

    /// Convert to a camera configuration
    ///
    /// ### Arguments
    ///
    /// - `camera` The camera
    ///
    /// ### Returns
    ///
    /// The camera configuration
    ///
    pub fn to_camera_config(&self, camera: &impl Camera) -> Result<CameraConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_to_camera_config(self.ptr, camera.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

    /// Get the camera operating environment
    ///
    /// ### Returns
    ///
    /// Imaging environment
    ///
    pub fn get_environment(&self) -> Result<ImagingEnvironment> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_environment(self.ptr, err.ptr()) };
        let out: ImagingEnvironment = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the camera operating environment
    ///
    /// ### Arguments
    ///
    /// - `imaging_environment` Imaging environment
    ///
    pub fn set_environment(&mut self, imaging_environment: ImagingEnvironment) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_environment(self.ptr, imaging_environment as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum distance photons reflect from
    ///
    /// ### Returns
    ///
    /// Maximum distance in meters
    ///
    pub fn get_max_distance(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_max_distance(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum distance photons reflect from
    ///
    /// ### Arguments
    ///
    /// - `max_distance` Maximum distance in meters
    ///
    pub fn set_max_distance(&mut self, max_distance: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_max_distance(self.ptr, max_distance, err.ptr()) };
        err.empty_result()
    }

    /// Get the depth frame rate
    ///
    /// ### Returns
    ///
    /// The depth fps
    ///
    pub fn get_fps(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_fps(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the depth frame rate
    ///
    /// ### Arguments
    ///
    /// - `fps` The depth fps
    ///
    pub fn set_fps(&mut self, fps: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_fps(self.ptr, fps, err.ptr()) };
        err.empty_result()
    }

    /// Get if lens flare is expected in the scene
    ///
    /// ### Returns
    ///
    /// if lens flare is present
    ///
    pub fn get_lens_flare(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_lens_flare(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set if lens flare is expected in the scene
    ///
    /// ### Arguments
    ///
    /// - `lens_flare` if lens flare is present
    ///
    pub fn set_lens_flare(&mut self, lens_flare: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_lens_flare(self.ptr, lens_flare, err.ptr()) };
        err.empty_result()
    }

    /// Get if there are translucent objects in the scene
    ///
    /// ### Returns
    ///
    /// If translucent objects
    ///
    pub fn get_translucent(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_translucent(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set if there are translucent objects in the scene
    ///
    /// ### Arguments
    ///
    /// - `translucent` If translucent objects
    ///
    pub fn set_translucent(&mut self, translucent: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_translucent(self.ptr, translucent, err.ptr()) };
        err.empty_result()
    }

    /// Get high dynamic range enabled. Enable this when pixel saturation
    /// occurs. This will only happen for retroreflectors or reflective objects
    /// close (<20cm) to the camera
    ///
    /// ### Returns
    ///
    /// Enable high dynamic range
    ///
    pub fn get_hdr(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_hdr(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set high dynamic range enabled. Enable this when pixel saturation
    /// occurs. This will only happen for retroreflectors or reflective objects
    /// close (<20cm) to the camera
    ///
    /// ### Arguments
    ///
    /// - `hdr` Enable high dynamic range
    ///
    pub fn set_hdr(&mut self, hdr: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_hdr(self.ptr, hdr, err.ptr()) };
        err.empty_result()
    }

    /// Get the integration time
    ///
    /// ### Returns
    ///
    /// The integration time
    ///
    pub fn get_integration_time(&self) -> Result<IntegrationTime> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_integration_time(self.ptr, err.ptr()) };
        let out: IntegrationTime = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the integration time
    ///
    /// ### Arguments
    ///
    /// - `int_time` The integration time
    ///
    pub fn set_integration_time(&mut self, int_time: IntegrationTime) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_integration_time(self.ptr, int_time as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get what the configuration is being optimized for
    ///
    /// ### Returns
    ///
    /// The user configuration strategy
    ///
    pub fn get_strategy(&self) -> Result<Strategy> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_strategy(self.ptr, err.ptr()) };
        let out: Strategy = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set what the configuration is being optimized for
    ///
    /// ### Arguments
    ///
    /// - `strategy` The user configuration strategy
    ///
    pub fn set_strategy(&mut self, strategy: Strategy) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_strategy(self.ptr, strategy as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get The camera channel, increment to support multi-camera usage
    ///
    /// ### Returns
    ///
    /// The channel number, set to 0, otherwise increment with number of
    /// cameras.
    ///
    pub fn get_channel(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_user_config_get_channel(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The camera channel, increment to support multi-camera usage
    ///
    /// ### Arguments
    ///
    /// - `channel` The channel number, set to 0, otherwise increment with
    ///
    pub fn set_channel(&mut self, channel: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_user_config_set_channel(self.ptr, channel, err.ptr()) };
        err.empty_result()
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_hdr_possible(_: *mut tof_camera_config, _: *mut *mut tof_error) -> bool;
    fn tof_phase_unwrap_possible(_: *mut tof_camera_config, _: *mut *mut tof_error) -> bool;
}

/// Check whether hdr is possible with current camera config
///
/// ### Arguments
///
/// - `config` The camera config to check
///
/// ### Returns
///
/// Whether hdr is possible
///
pub fn hdr_possible(config: &CameraConfig) -> Result<bool> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_hdr_possible(config.ptr(), err.ptr()) };
    err.make_result(ret)
}

/// Check whether phase unwrap is possible with current camera config
///
/// ### Arguments
///
/// - `config` The camera config to check
///
/// ### Returns
///
/// Whether phase unwrap is possible
///
pub fn phase_unwrap_possible(config: &CameraConfig) -> Result<bool> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_phase_unwrap_possible(config.ptr(), err.ptr()) };
    err.make_result(ret)
}

/// The shape of which surrounding pixels are checked in the flying pixel filter
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum FlyingShape {
    /// Four neighboring pixels in a plus shape
    Plus = 0,
    /// All eight neighboring pixels
    Box = 1,
}

/// Mixed pixel sort mode. The mode you want to use depends on your application.
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum MpSortMode {
    /// Choose darkest path
    Darkest = 0,
    /// Return the surface of a translucent object, resolve corners or
    /// checkerboard lens flare
    Shortest = 1,
    /// Mixed pixels between foreground and background, get the best SNR
    Brightest = 2,
    /// Correct lens flare from close bright objects
    Longest = 3,
    /// Correct lens flare
    Aggressor = 4,
    /// Adaptive shortest
    AdaptiveShortV0 = 5,
    /// Adaptive shortest version two
    AdaptiveShortV1 = 6,
    /// Correct lens flare version two
    AggressorV2 = 7,
}

/// What kind of binning to do
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum BinningMode {
    /// Normal binning, a BxB grid averaged to a single pixel
    Normal = 0,
    /// Smart binning, take into account the standard deviation around a chosen
    /// pixel. Similar to local means but reduces the image size
    Smart = 1,
}

/// In what dimension the xyz is exported
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum XyzDimension {
    /// Export in millimeters
    Millimeter = 0,
    /// Export in meters
    Meter = 1,
}

#[repr(C)]
struct tof_processing_config {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_processing_config_delete(_: *mut tof_processing_config);
    fn tof_processing_config_new_default(_: *mut *mut tof_error) -> *mut tof_processing_config;
    fn tof_processing_config_new_from_disk(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_processing_config;
    fn tof_processing_config_write(_: *const tof_processing_config, _: *const c_char, _: *mut *mut tof_error);
    fn tof_processing_config_get_gpu(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_gpu(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_calibration_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_calibration_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_phase_unwrapping_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_phase_unwrapping_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_phase_unwrapping_max_offset(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_phase_unwrapping_max_offset(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mpi_detect_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_mpi_detect_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_mpi_detect_aratio(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_mpi_detect_aratio(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_mpi_detect_phi_diff(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_mpi_detect_phi_diff(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_mixed_pixel_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_lut_file(_: *const tof_processing_config, _: *mut *mut tof_error) -> *const c_char;
    fn tof_processing_config_set_mixed_pixel_lut_file(_: *mut tof_processing_config, _: *const c_char, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_amp_min(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_amp_min(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_sort_mode(_: *const tof_processing_config, _: *mut *mut tof_error) -> i32;
    fn tof_processing_config_set_mixed_pixel_sort_mode(_: *mut tof_processing_config, _: i32, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_aratio(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_aratio(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_phi_diff(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_phi_diff(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_aggressor_threshold(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_aggressor_threshold(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_min_aggressor_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> i32;
    fn tof_processing_config_set_mixed_pixel_min_aggressor_size(_: *mut tof_processing_config, _: i32, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_max_error_3f(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_max_error_3f(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_adaptive_amp_max(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_adaptive_amp_max(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_adaptive_amp_min(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_adaptive_amp_min(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_aratio_min(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_aratio_min(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_mixed_pixel_aggressor_phi_offset(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_mixed_pixel_aggressor_phi_offset(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_hdr_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_hdr_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_average_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_average_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_average_nframes(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_average_nframes(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_temporal_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_temporal_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_temporal_sigma(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_temporal_sigma(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_temporal_nframes(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_temporal_nframes(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_gaussian_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_gaussian_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_gaussian_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_gaussian_size(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_gaussian_sigma(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_gaussian_sigma(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_median_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_median_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_median_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_median_size(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_bilateral_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_bilateral_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_bilateral_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_bilateral_size(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_bilateral_sigma(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_bilateral_sigma(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_local_means_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_local_means_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_local_means_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_local_means_size(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_local_means_sigma(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_local_means_sigma(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_flying_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_flying_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_flying_shape(_: *const tof_processing_config, _: *mut *mut tof_error) -> i32;
    fn tof_processing_config_set_flying_shape(_: *mut tof_processing_config, _: i32, _: *mut *mut tof_error);
    fn tof_processing_config_get_flying_distance(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_flying_distance(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_flying_edges(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_flying_edges(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_amp_threshold_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_amp_threshold_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_amp_threshold_min(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_amp_threshold_min(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_amp_threshold_max(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_amp_threshold_max(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_dist_threshold_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_dist_threshold_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_dist_threshold_min(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_dist_threshold_min(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_dist_threshold_max(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_dist_threshold_max(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_radial_scale(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_radial_scale(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_radial_add(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_radial_add(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_intensity_scale(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_intensity_scale(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_intensity_add(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_intensity_add(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_char_add(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_char_add(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_char_mult(_: *const tof_processing_config, _: *mut *mut tof_error) -> f64;
    fn tof_processing_config_set_char_mult(_: *mut tof_processing_config, _: f64, _: *mut *mut tof_error);
    fn tof_processing_config_get_binning_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_binning_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_binning_size(_: *const tof_processing_config, _: *mut *mut tof_error) -> usize;
    fn tof_processing_config_set_binning_size(_: *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
    fn tof_processing_config_get_binning_mode(_: *const tof_processing_config, _: *mut *mut tof_error) -> i32;
    fn tof_processing_config_set_binning_mode(_: *mut tof_processing_config, _: i32, _: *mut *mut tof_error);
    fn tof_processing_config_get_binning_sigma(_: *const tof_processing_config, _: *mut *mut tof_error) -> f32;
    fn tof_processing_config_set_binning_sigma(_: *mut tof_processing_config, _: f32, _: *mut *mut tof_error);
    fn tof_processing_config_get_xyz_dimension(_: *const tof_processing_config, _: *mut *mut tof_error) -> i32;
    fn tof_processing_config_set_xyz_dimension(_: *mut tof_processing_config, _: i32, _: *mut *mut tof_error);
    fn tof_processing_config_get_rigid_transformation_enabled(_: *const tof_processing_config, _: *mut *mut tof_error) -> bool;
    fn tof_processing_config_set_rigid_transformation_enabled(_: *mut tof_processing_config, _: bool, _: *mut *mut tof_error);
    fn tof_processing_config_get_rigid_transformation_matrix(_: *const tof_processing_config, _: *mut *mut tof_error) -> *const f32;
    fn tof_processing_config_set_rigid_transformation_matrix(_: *mut tof_processing_config, _: *const f32, _: *mut *mut tof_error);
}

/// Processing that can be done
pub struct ProcessingConfig {
    ptr: *mut tof_processing_config,
}

impl Drop for ProcessingConfig {
    fn drop(&mut self) {
        unsafe { tof_processing_config_delete(self.ptr); }
    }
}

unsafe impl Send for ProcessingConfig {}

impl ProcessingConfig {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_processing_config {
        self.ptr
    }

    /// Create default processing config
    ///
    pub fn new_default() -> Result<Self> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_new_default(err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(ProcessingConfig { ptr: ret })
        }
    }

    /// Load processing config from disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location of processing config
    ///
    pub fn new_from_disk(file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_processing_config_new_from_disk(file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(ProcessingConfig { ptr: ret })
        }
    }

    /// Write processing config to disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location where to save the processing config
    ///
    pub fn write(&self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_processing_config_write(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get whether processing on the gpu is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_gpu(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_gpu(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether processing on the gpu is enabled
    ///
    /// ### Arguments
    ///
    /// - `gpu` Enabled
    ///
    pub fn set_gpu(&mut self, gpu: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_gpu(self.ptr, gpu, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the calibration is applied
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_calibration_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_calibration_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the calibration is applied
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_calibration_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_calibration_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get whether phase unwrapping is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_phase_unwrapping_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_phase_unwrapping_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether phase unwrapping is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_phase_unwrapping_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_phase_unwrapping_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get The maximum distance between ideal location and actual
    ///
    /// ### Returns
    ///
    /// Max offset, range [0, 1]
    ///
    pub fn get_phase_unwrapping_max_offset(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_phase_unwrapping_max_offset(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The maximum distance between ideal location and actual
    ///
    /// ### Arguments
    ///
    /// - `max_offset` Max offset, range [0, 1]
    ///
    pub fn set_phase_unwrapping_max_offset(&mut self, max_offset: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_phase_unwrapping_max_offset(self.ptr, max_offset, err.ptr()) };
        err.empty_result()
    }

    /// Get wether mpi detect is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_mpi_detect_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mpi_detect_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set wether mpi detect is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_mpi_detect_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mpi_detect_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get The amplitude ratio value
    ///
    /// ### Returns
    ///
    /// The maximum ratio
    ///
    pub fn get_mpi_detect_aratio(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mpi_detect_aratio(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The amplitude ratio value
    ///
    /// ### Arguments
    ///
    /// - `aratio` The maximum ratio
    ///
    pub fn set_mpi_detect_aratio(&mut self, aratio: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mpi_detect_aratio(self.ptr, aratio, err.ptr()) };
        err.empty_result()
    }

    /// Get the phase difference ratio
    ///
    /// ### Returns
    ///
    /// The maximum phase difference
    ///
    pub fn get_mpi_detect_phi_diff(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mpi_detect_phi_diff(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the phase difference ratio
    ///
    /// ### Arguments
    ///
    /// - `phi_diff` The maximum phase difference
    ///
    pub fn set_mpi_detect_phi_diff(&mut self, phi_diff: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mpi_detect_phi_diff(self.ptr, phi_diff, err.ptr()) };
        err.empty_result()
    }

    /// Get wether mixed pixel is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_mixed_pixel_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set wether mixed pixel is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_mixed_pixel_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get location of the mixed pixel look up table file
    ///
    /// ### Returns
    ///
    /// File location
    ///
    pub fn get_mixed_pixel_lut_file(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_lut_file(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Set location of the mixed pixel look up table file
    ///
    /// ### Arguments
    ///
    /// - `file_location` File location
    ///
    pub fn set_mixed_pixel_lut_file(&mut self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_processing_config_set_mixed_pixel_lut_file(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get minimum amplitude that pixel needs to have to apply mixed pixel to
    /// it
    ///
    /// ### Returns
    ///
    /// Minimum amplitude
    ///
    pub fn get_mixed_pixel_amp_min(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_amp_min(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set minimum amplitude that pixel needs to have to apply mixed pixel to
    /// it
    ///
    /// ### Arguments
    ///
    /// - `amp_min` Minimum amplitude
    ///
    pub fn set_mixed_pixel_amp_min(&mut self, amp_min: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_amp_min(self.ptr, amp_min, err.ptr()) };
        err.empty_result()
    }

    /// Get what output is selected
    ///
    /// ### Returns
    ///
    /// Sort mode
    ///
    pub fn get_mixed_pixel_sort_mode(&self) -> Result<MpSortMode> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_sort_mode(self.ptr, err.ptr()) };
        let out: MpSortMode = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set what output is selected
    ///
    /// ### Arguments
    ///
    /// - `sort_mode` Sort mode
    ///
    pub fn set_mixed_pixel_sort_mode(&mut self, sort_mode: MpSortMode) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_sort_mode(self.ptr, sort_mode as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get the amplitude ratio for detecting MPI
    ///
    /// ### Returns
    ///
    /// Amplitude ratio
    ///
    pub fn get_mixed_pixel_aratio(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_aratio(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the amplitude ratio for detecting MPI
    ///
    /// ### Arguments
    ///
    /// - `aratio` Amplitude ratio
    ///
    pub fn set_mixed_pixel_aratio(&mut self, aratio: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_aratio(self.ptr, aratio, err.ptr()) };
        err.empty_result()
    }

    /// Get the phase difference for detecting MPI
    ///
    /// ### Returns
    ///
    /// Phase difference
    ///
    pub fn get_mixed_pixel_phi_diff(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_phi_diff(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the phase difference for detecting MPI
    ///
    /// ### Arguments
    ///
    /// - `phi_diff` Phase difference
    ///
    pub fn set_mixed_pixel_phi_diff(&mut self, phi_diff: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_phi_diff(self.ptr, phi_diff, err.ptr()) };
        err.empty_result()
    }

    /// Get the threshold to classify a pixel as an aggressor
    ///
    /// ### Returns
    ///
    /// Aggressor threshold
    ///
    pub fn get_mixed_pixel_aggressor_threshold(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_aggressor_threshold(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the threshold to classify a pixel as an aggressor
    ///
    /// ### Arguments
    ///
    /// - `threshold` Aggressor threshold
    ///
    pub fn set_mixed_pixel_aggressor_threshold(&mut self, threshold: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_aggressor_threshold(self.ptr, threshold, err.ptr()) };
        err.empty_result()
    }

    /// Get the minimum amount of bright pixels before a bright spot is
    /// classified as an aggressor
    ///
    /// ### Returns
    ///
    /// Minimum aggressor size
    ///
    pub fn get_mixed_pixel_min_aggressor_size(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_min_aggressor_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the minimum amount of bright pixels before a bright spot is
    /// classified as an aggressor
    ///
    /// ### Arguments
    ///
    /// - `size` Minimum aggressor size
    ///
    pub fn set_mixed_pixel_min_aggressor_size(&mut self, size: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_min_aggressor_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum error when using three modulation frequencies
    ///
    /// ### Returns
    ///
    /// The error
    ///
    pub fn get_mixed_pixel_max_error_3f(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_max_error_3f(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum error when using three modulation frequencies
    ///
    /// ### Arguments
    ///
    /// - `max_error_3f` The error
    ///
    pub fn set_mixed_pixel_max_error_3f(&mut self, max_error_3f: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_max_error_3f(self.ptr, max_error_3f, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum amplitude for adaptive sorting
    ///
    /// ### Returns
    ///
    /// Maximum amplitude
    ///
    pub fn get_mixed_pixel_adaptive_amp_max(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_adaptive_amp_max(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum amplitude for adaptive sorting
    ///
    /// ### Arguments
    ///
    /// - `adaptive_amp_max` Maximum amplitude
    ///
    pub fn set_mixed_pixel_adaptive_amp_max(&mut self, adaptive_amp_max: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_adaptive_amp_max(self.ptr, adaptive_amp_max, err.ptr()) };
        err.empty_result()
    }

    /// Get the minimum amplitude for adaptive sorting
    ///
    /// ### Returns
    ///
    /// Minimum amplitude
    ///
    pub fn get_mixed_pixel_adaptive_amp_min(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_adaptive_amp_min(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the minimum amplitude for adaptive sorting
    ///
    /// ### Arguments
    ///
    /// - `adaptive_amp_min` Minimum amplitude
    ///
    pub fn set_mixed_pixel_adaptive_amp_min(&mut self, adaptive_amp_min: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_adaptive_amp_min(self.ptr, adaptive_amp_min, err.ptr()) };
        err.empty_result()
    }

    /// Get the minimum return amplitude ratio for adaptive sorting
    ///
    /// ### Returns
    ///
    /// Minimum return ratio
    ///
    pub fn get_mixed_pixel_aratio_min(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_aratio_min(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the minimum return amplitude ratio for adaptive sorting
    ///
    /// ### Arguments
    ///
    /// - `aratio_min` Minimum return ratio
    ///
    pub fn set_mixed_pixel_aratio_min(&mut self, aratio_min: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_aratio_min(self.ptr, aratio_min, err.ptr()) };
        err.empty_result()
    }

    /// Get The maximum phase difference before a return is not classified as
    /// aggressor
    ///
    /// ### Returns
    ///
    /// Phase offset before not a aggressor return
    ///
    pub fn get_mixed_pixel_aggressor_phi_offset(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_mixed_pixel_aggressor_phi_offset(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The maximum phase difference before a return is not classified as
    /// aggressor
    ///
    /// ### Arguments
    ///
    /// - `aggressor_phi_offset` Phase offset before not a aggressor return
    ///
    pub fn set_mixed_pixel_aggressor_phi_offset(&mut self, aggressor_phi_offset: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_mixed_pixel_aggressor_phi_offset(self.ptr, aggressor_phi_offset, err.ptr()) };
        err.empty_result()
    }

    /// Get whether hdr is enabled. High dynamic range allows you to separate
    /// depth with high intensity difference. This is mostly useful when the
    /// sensor is being saturated.
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_hdr_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_hdr_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether hdr is enabled. High dynamic range allows you to separate
    /// depth with high intensity difference. This is mostly useful when the
    /// sensor is being saturated.
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_hdr_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_hdr_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get whether averaging is enabled. The average process allows you to
    /// average multiple frames together at the cost of frame rate
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_average_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_average_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether averaging is enabled. The average process allows you to
    /// average multiple frames together at the cost of frame rate
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_average_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_average_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the number of frames to average together
    ///
    /// ### Returns
    ///
    /// Number of frames
    ///
    pub fn get_average_nframes(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_average_nframes(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the number of frames to average together
    ///
    /// ### Arguments
    ///
    /// - `nframes` Number of frames
    ///
    pub fn set_average_nframes(&mut self, nframes: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_average_nframes(self.ptr, nframes, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the temporal filter is enabled. The temporal filter is a
    /// running average as long as the values are within sigma standard
    /// deviations of each other
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_temporal_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_temporal_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the temporal filter is enabled. The temporal filter is a
    /// running average as long as the values are within sigma standard
    /// deviations of each other
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_temporal_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_temporal_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the number of standard deviations before resetting the running
    /// average
    ///
    /// ### Returns
    ///
    /// The standard deviation
    ///
    pub fn get_temporal_sigma(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_temporal_sigma(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the number of standard deviations before resetting the running
    /// average
    ///
    /// ### Arguments
    ///
    /// - `sigma` The standard deviation
    ///
    pub fn set_temporal_sigma(&mut self, sigma: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_temporal_sigma(self.ptr, sigma, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum number of frames to average together
    ///
    /// ### Returns
    ///
    /// Number of frames
    ///
    pub fn get_temporal_nframes(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_temporal_nframes(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum number of frames to average together
    ///
    /// ### Arguments
    ///
    /// - `nframes` Number of frames
    ///
    pub fn set_temporal_nframes(&mut self, nframes: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_temporal_nframes(self.ptr, nframes, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the gaussian blur is enabled. This is a standard gaussian
    /// blur applied over the depth image
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_gaussian_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_gaussian_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the gaussian blur is enabled. This is a standard gaussian
    /// blur applied over the depth image
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_gaussian_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_gaussian_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the gaussian region size
    ///
    /// ### Returns
    ///
    /// Region size, 3x3 and 5x5 are supported
    ///
    pub fn get_gaussian_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_gaussian_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the gaussian region size
    ///
    /// ### Arguments
    ///
    /// - `size` Region size, 3x3 and 5x5 are supported
    ///
    pub fn set_gaussian_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_gaussian_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get the gaussian sigma
    ///
    /// ### Returns
    ///
    /// The standard deviation
    ///
    pub fn get_gaussian_sigma(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_gaussian_sigma(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the gaussian sigma
    ///
    /// ### Arguments
    ///
    /// - `sigma` The standard deviation
    ///
    pub fn set_gaussian_sigma(&mut self, sigma: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_gaussian_sigma(self.ptr, sigma, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the median filter is enabled. This is a standard median
    /// filter applied over the depth image
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_median_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_median_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the median filter is enabled. This is a standard median
    /// filter applied over the depth image
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_median_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_median_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the median region size
    ///
    /// ### Returns
    ///
    /// Region size, 3x3 and 5x5 are supported
    ///
    pub fn get_median_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_median_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the median region size
    ///
    /// ### Arguments
    ///
    /// - `size` Region size, 3x3 and 5x5 are supported
    ///
    pub fn set_median_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_median_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the bilateral filter is enabled. This filter averages the
    /// current pixel with neighboring ones if they're within the specified
    /// standard deviation
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_bilateral_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_bilateral_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the bilateral filter is enabled. This filter averages the
    /// current pixel with neighboring ones if they're within the specified
    /// standard deviation
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_bilateral_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_bilateral_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the bilateral region size
    ///
    /// ### Returns
    ///
    /// Region size, 3x3 and 5x5 are supported
    ///
    pub fn get_bilateral_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_bilateral_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the bilateral region size
    ///
    /// ### Arguments
    ///
    /// - `size` Region size, 3x3 and 5x5 are supported
    ///
    pub fn set_bilateral_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_bilateral_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get the bilateral standard deviation
    ///
    /// ### Returns
    ///
    /// The standard deviation
    ///
    pub fn get_bilateral_sigma(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_bilateral_sigma(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the bilateral standard deviation
    ///
    /// ### Arguments
    ///
    /// - `sigma` The standard deviation
    ///
    pub fn set_bilateral_sigma(&mut self, sigma: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_bilateral_sigma(self.ptr, sigma, err.ptr()) };
        err.empty_result()
    }

    /// Get Whether the local means filter is enabled.
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_local_means_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_local_means_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set Whether the local means filter is enabled.
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_local_means_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_local_means_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get The local means region size
    ///
    /// ### Returns
    ///
    /// Region size
    ///
    pub fn get_local_means_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_local_means_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The local means region size
    ///
    /// ### Arguments
    ///
    /// - `size` Region size
    ///
    pub fn set_local_means_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_local_means_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get The local means standard deviation for how close values are
    ///
    /// ### Returns
    ///
    /// The stand deviation
    ///
    pub fn get_local_means_sigma(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_local_means_sigma(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The local means standard deviation for how close values are
    ///
    /// ### Arguments
    ///
    /// - `sigma` The stand deviation
    ///
    pub fn set_local_means_sigma(&mut self, sigma: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_local_means_sigma(self.ptr, sigma, err.ptr()) };
        err.empty_result()
    }

    /// Get whether flying pixel filter is enabled. The flying pixel filters out
    /// pixels when more than x edges are farther than x size away
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_flying_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_flying_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether flying pixel filter is enabled. The flying pixel filters out
    /// pixels when more than x edges are farther than x size away
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_flying_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_flying_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the shape of which surrounding pixels are checked
    ///
    /// ### Returns
    ///
    /// The shape
    ///
    pub fn get_flying_shape(&self) -> Result<FlyingShape> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_flying_shape(self.ptr, err.ptr()) };
        let out: FlyingShape = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the shape of which surrounding pixels are checked
    ///
    /// ### Arguments
    ///
    /// - `shape` The shape
    ///
    pub fn set_flying_shape(&mut self, shape: FlyingShape) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_flying_shape(self.ptr, shape as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum distance allowed between surrounding pixels
    ///
    /// ### Returns
    ///
    /// Maximum distance
    ///
    pub fn get_flying_distance(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_flying_distance(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum distance allowed between surrounding pixels
    ///
    /// ### Arguments
    ///
    /// - `distance` Maximum distance
    ///
    pub fn set_flying_distance(&mut self, distance: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_flying_distance(self.ptr, distance, err.ptr()) };
        err.empty_result()
    }

    /// Get the number of edges allowed to be larger than maximum distance
    ///
    /// ### Returns
    ///
    /// Maximum edges
    ///
    pub fn get_flying_edges(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_flying_edges(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the number of edges allowed to be larger than maximum distance
    ///
    /// ### Arguments
    ///
    /// - `edges` Maximum edges
    ///
    pub fn set_flying_edges(&mut self, edges: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_flying_edges(self.ptr, edges, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the amplitude threshold filter is enabled. This filter
    /// filters out pixels outside of the specified values
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_amp_threshold_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_amp_threshold_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the amplitude threshold filter is enabled. This filter
    /// filters out pixels outside of the specified values
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_amp_threshold_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_amp_threshold_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the minimum amplitude allowed
    ///
    /// ### Returns
    ///
    /// Minimum value
    ///
    pub fn get_amp_threshold_min(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_amp_threshold_min(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the minimum amplitude allowed
    ///
    /// ### Arguments
    ///
    /// - `min` Minimum value
    ///
    pub fn set_amp_threshold_min(&mut self, min: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_amp_threshold_min(self.ptr, min, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum amplitude allowed
    ///
    /// ### Returns
    ///
    /// Maximum value
    ///
    pub fn get_amp_threshold_max(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_amp_threshold_max(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum amplitude allowed
    ///
    /// ### Arguments
    ///
    /// - `max` Maximum value
    ///
    pub fn set_amp_threshold_max(&mut self, max: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_amp_threshold_max(self.ptr, max, err.ptr()) };
        err.empty_result()
    }

    /// Get whether the distance threshold filter is enabled. This filter
    /// filters out pixels outside of the specified values
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_dist_threshold_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_dist_threshold_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether the distance threshold filter is enabled. This filter
    /// filters out pixels outside of the specified values
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_dist_threshold_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_dist_threshold_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the minimum distance allowed
    ///
    /// ### Returns
    ///
    /// Minimum value
    ///
    pub fn get_dist_threshold_min(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_dist_threshold_min(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the minimum distance allowed
    ///
    /// ### Arguments
    ///
    /// - `min` Minimum value
    ///
    pub fn set_dist_threshold_min(&mut self, min: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_dist_threshold_min(self.ptr, min, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum distance allowed
    ///
    /// ### Returns
    ///
    /// Maximum value
    ///
    pub fn get_dist_threshold_max(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_dist_threshold_max(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the maximum distance allowed
    ///
    /// ### Arguments
    ///
    /// - `max` Maximum value
    ///
    pub fn set_dist_threshold_max(&mut self, max: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_dist_threshold_max(self.ptr, max, err.ptr()) };
        err.empty_result()
    }

    /// Get how much to scale the radial value by
    ///
    /// ### Returns
    ///
    /// Radial scale
    ///
    pub fn get_radial_scale(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_radial_scale(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set how much to scale the radial value by
    ///
    /// ### Arguments
    ///
    /// - `scale` Radial scale
    ///
    pub fn set_radial_scale(&mut self, scale: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_radial_scale(self.ptr, scale, err.ptr()) };
        err.empty_result()
    }

    /// Get how much to add to the radial value
    ///
    /// ### Returns
    ///
    /// Radial addition
    ///
    pub fn get_radial_add(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_radial_add(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set how much to add to the radial value
    ///
    /// ### Arguments
    ///
    /// - `addition` Radial addition
    ///
    pub fn set_radial_add(&mut self, addition: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_radial_add(self.ptr, addition, err.ptr()) };
        err.empty_result()
    }

    /// Get how much to scale the intensity value by
    ///
    /// ### Returns
    ///
    /// Intensity scale
    ///
    pub fn get_intensity_scale(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_intensity_scale(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set how much to scale the intensity value by
    ///
    /// ### Arguments
    ///
    /// - `scale` Intensity scale
    ///
    pub fn set_intensity_scale(&mut self, scale: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_intensity_scale(self.ptr, scale, err.ptr()) };
        err.empty_result()
    }

    /// Get how much to add to the intensity value
    ///
    /// ### Returns
    ///
    /// Intensity addition
    ///
    pub fn get_intensity_add(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_intensity_add(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set how much to add to the intensity value
    ///
    /// ### Arguments
    ///
    /// - `addition` Intensity addition
    ///
    pub fn set_intensity_add(&mut self, addition: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_intensity_add(self.ptr, addition, err.ptr()) };
        err.empty_result()
    }

    /// Get the characterization addition component. The characterization
    /// function computes the sigma for the temporal and bilateral filter
    ///
    /// ### Returns
    ///
    /// addition
    ///
    pub fn get_char_add(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_char_add(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the characterization addition component. The characterization
    /// function computes the sigma for the temporal and bilateral filter
    ///
    /// ### Arguments
    ///
    /// - `addition` addition
    ///
    pub fn set_char_add(&mut self, addition: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_char_add(self.ptr, addition, err.ptr()) };
        err.empty_result()
    }

    /// Get the characterization multiplication component. The characterization
    /// function computes the sigma for the temporal and bilateral filter
    ///
    /// ### Returns
    ///
    /// Multiplication
    ///
    pub fn get_char_mult(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_char_mult(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the characterization multiplication component. The characterization
    /// function computes the sigma for the temporal and bilateral filter
    ///
    /// ### Arguments
    ///
    /// - `multiplication` Multiplication
    ///
    pub fn set_char_mult(&mut self, multiplication: f64) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_char_mult(self.ptr, multiplication, err.ptr()) };
        err.empty_result()
    }

    /// Get whether software binning is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_binning_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_binning_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether software binning is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_binning_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_binning_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the size of the area to bin, 2 would bin a 2x2 area
    ///
    /// ### Returns
    ///
    /// The bin size
    ///
    pub fn get_binning_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_binning_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the size of the area to bin, 2 would bin a 2x2 area
    ///
    /// ### Arguments
    ///
    /// - `size` The bin size
    ///
    pub fn set_binning_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_binning_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Get the binning mode to use
    ///
    /// ### Returns
    ///
    /// The binning mode
    ///
    pub fn get_binning_mode(&self) -> Result<BinningMode> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_binning_mode(self.ptr, err.ptr()) };
        let out: BinningMode = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the binning mode to use
    ///
    /// ### Arguments
    ///
    /// - `mode` The binning mode
    ///
    pub fn set_binning_mode(&mut self, mode: BinningMode) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_binning_mode(self.ptr, mode as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get The maximum standard deviation to use, only used in the smart
    /// binning mode
    ///
    /// ### Returns
    ///
    /// The standard deviation
    ///
    pub fn get_binning_sigma(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_binning_sigma(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set The maximum standard deviation to use, only used in the smart
    /// binning mode
    ///
    /// ### Arguments
    ///
    /// - `sigma` The standard deviation
    ///
    pub fn set_binning_sigma(&mut self, sigma: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_binning_sigma(self.ptr, sigma, err.ptr()) };
        err.empty_result()
    }

    /// Get the dimension of the xyz output
    ///
    /// ### Returns
    ///
    /// The dimension
    ///
    pub fn get_xyz_dimension(&self) -> Result<XyzDimension> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_xyz_dimension(self.ptr, err.ptr()) };
        let out: XyzDimension = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the dimension of the xyz output
    ///
    /// ### Arguments
    ///
    /// - `dimension` The dimension
    ///
    pub fn set_xyz_dimension(&mut self, dimension: XyzDimension) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_xyz_dimension(self.ptr, dimension as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get whether rigid transformation is enabled
    ///
    /// ### Returns
    ///
    /// Enabled
    ///
    pub fn get_rigid_transformation_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_rigid_transformation_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set whether rigid transformation is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Enabled
    ///
    pub fn set_rigid_transformation_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_rigid_transformation_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

    /// Get the rigid transformation 4x4 matrix
    ///
    /// ### Returns
    ///
    /// 4x4 transformation matrix
    ///
    pub fn get_rigid_transformation_matrix(&self) -> Result<[f32; 16]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_processing_config_get_rigid_transformation_matrix(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f32; 16]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 16);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Set the rigid transformation 4x4 matrix
    ///
    /// ### Arguments
    ///
    /// - `matrix` 4x4 transformation matrix
    ///
    pub fn set_rigid_transformation_matrix(&mut self, matrix: [f32; 16]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_processing_config_set_rigid_transformation_matrix(self.ptr, matrix.as_ptr(), err.ptr()) };
        err.empty_result()
    }

}

/// ToF cameras made by Chronoptics
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum CameraType {
    /// Original Moa and first generation Kea camera with opt sensor
    Opt8241 = 1,
    /// First generation Kea camera with opt sensor and external modulation
    Opt8241Ext = 4,
    /// Newer generation Kea cameras with mlx sensor
    Mlx75027 = 6,
    /// Kea st camera
    Vd55h1 = 9,
}

/// Operating mode of the cameras' sensor
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum SensorMode {
    /// Normal mode ToF sensor mod (A - B)
    Normal = 0,
    /// Common mode (A + B)
    Common = 1,
    /// Use external demodulation
    ExternalDemod = 4,
    /// Use the light source as an input
    ExternalSignal = 5,
    /// Only tap a is read out
    TapA = 8,
    /// Only tap b is read out
    TapB = 9,
    /// Both A and B tap are read out, this means double the data transmission
    TapAb = 10,
    /// Both A and B tap are read out for both a short and long integration time
    TapAbHdr = 12,
}

/// The trigger mode of the camera
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum TriggerMode {
    /// The camera is continuously running and outputting frames
    Continuous = 0,
    /// The camera waits for a hardware trigger
    Hardware = 1,
    /// The camera waits for a software trigger
    Software = 2,
}

#[repr(C)]
struct tof_roi {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_roi_delete(_: *mut tof_roi);
    fn tof_roi_sensor_rows(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_sensor_cols(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_get_row_offset(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_set_row_offset(_: *mut tof_roi, _: i32, _: *mut *mut tof_error);
    fn tof_roi_get_col_offset(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_set_col_offset(_: *mut tof_roi, _: i32, _: *mut *mut tof_error);
    fn tof_roi_get_img_rows(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_set_img_rows(_: *mut tof_roi, _: i32, _: *mut *mut tof_error);
    fn tof_roi_get_img_cols(_: *const tof_roi, _: *mut *mut tof_error) -> i32;
    fn tof_roi_set_img_cols(_: *mut tof_roi, _: i32, _: *mut *mut tof_error);
}

/// Region of interest
pub struct Roi {
    ptr: *mut tof_roi,
}

impl Drop for Roi {
    fn drop(&mut self) {
        unsafe { tof_roi_delete(self.ptr); }
    }
}

unsafe impl Send for Roi {}

impl Roi {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_roi {
        self.ptr
    }

    /// Get the number of rows in the image sensor
    ///
    /// ### Returns
    ///
    /// Number of rows
    ///
    pub fn sensor_rows(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_sensor_rows(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the number of columns in the image sensor
    ///
    /// ### Returns
    ///
    /// Number of columsn
    ///
    pub fn sensor_cols(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_sensor_cols(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the row offset in pixels
    ///
    /// ### Returns
    ///
    /// Row offset
    ///
    pub fn get_row_offset(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_get_row_offset(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the row offset in pixels
    ///
    /// ### Arguments
    ///
    /// - `row_offset` Row offset
    ///
    pub fn set_row_offset(&mut self, row_offset: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_roi_set_row_offset(self.ptr, row_offset, err.ptr()) };
        err.empty_result()
    }

    /// Get the column offset in pixels
    ///
    /// ### Returns
    ///
    /// Column offset
    ///
    pub fn get_col_offset(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_get_col_offset(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the column offset in pixels
    ///
    /// ### Arguments
    ///
    /// - `col_offset` Column offset
    ///
    pub fn set_col_offset(&mut self, col_offset: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_roi_set_col_offset(self.ptr, col_offset, err.ptr()) };
        err.empty_result()
    }

    /// Get the number of rows in the output image
    ///
    /// ### Returns
    ///
    /// Image rows
    ///
    pub fn get_img_rows(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_get_img_rows(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the number of rows in the output image
    ///
    /// ### Arguments
    ///
    /// - `img_rows` Image rows
    ///
    pub fn set_img_rows(&mut self, img_rows: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_roi_set_img_rows(self.ptr, img_rows, err.ptr()) };
        err.empty_result()
    }

    /// Get the number of columns in the output image
    ///
    /// ### Returns
    ///
    /// Image columns
    ///
    pub fn get_img_cols(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_roi_get_img_cols(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the number of columns in the output image
    ///
    /// ### Arguments
    ///
    /// - `img_cols` Image columns
    ///
    pub fn set_img_cols(&mut self, img_cols: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_roi_set_img_cols(self.ptr, img_cols, err.ptr()) };
        err.empty_result()
    }

}

#[repr(C)]
struct tof_camera_config {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_camera_config_delete(_: *mut tof_camera_config);
    fn tof_camera_config_new_from_disk(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_camera_config;
    fn tof_camera_config_write(_: *const tof_camera_config, _: *const c_char, _: *mut *mut tof_error);
    fn tof_camera_config_get_type(_: *const tof_camera_config, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_reset(_: *mut tof_camera_config, _: *mut *mut tof_error);
    fn tof_camera_config_frame_size(_: *const tof_camera_config, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_add_frame(_: *mut tof_camera_config, _: *mut *mut tof_error);
    fn tof_camera_config_default_processing(_: *const tof_camera_config, _: *mut *mut tof_error) -> *mut tof_processing_config;
    fn tof_camera_config_erase_frame(_: *mut tof_camera_config, _: usize, _: *mut *mut tof_error);
    fn tof_camera_config_get_phase_shifts(_: *const tof_camera_config, _: usize, _: *const f32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_set_phase_shifts(_: *mut tof_camera_config, _: usize, _: *const f32, _: usize, _: *mut *mut tof_error);
    fn tof_camera_config_get_integration_time(_: *const tof_camera_config, _: usize, _: *const u32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_set_integration_time(_: *mut tof_camera_config, _: usize, _: *const u32, _: usize, _: *mut *mut tof_error);
    fn tof_camera_config_get_sensor_mode(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_set_sensor_mode(_: *mut tof_camera_config, _: usize, _: i32, _: *mut *mut tof_error);
    fn tof_camera_config_get_modulation_frequency(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> f32;
    fn tof_camera_config_set_modulation_frequency(_: *mut tof_camera_config, _: usize, _: f32, _: *mut *mut tof_error);
    fn tof_camera_config_get_duty_cycle(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> f32;
    fn tof_camera_config_set_duty_cycle(_: *mut tof_camera_config, _: usize, _: f32, _: *mut *mut tof_error);
    fn tof_camera_config_get_roi(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> *mut tof_roi;
    fn tof_camera_config_set_roi(_: *mut tof_camera_config, _: usize, _: *mut tof_roi, _: *mut *mut tof_error);
    fn tof_camera_config_get_binning(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> u8;
    fn tof_camera_config_set_binning(_: *mut tof_camera_config, _: usize, _: u8, _: *mut *mut tof_error);
    fn tof_camera_config_get_flip(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> bool;
    fn tof_camera_config_set_flip(_: *mut tof_camera_config, _: usize, _: bool, _: *mut *mut tof_error);
    fn tof_camera_config_get_mirror(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> bool;
    fn tof_camera_config_set_mirror(_: *mut tof_camera_config, _: usize, _: bool, _: *mut *mut tof_error);
    fn tof_camera_config_get_gain(_: *const tof_camera_config, _: *mut *mut tof_error) -> f32;
    fn tof_camera_config_set_gain(_: *mut tof_camera_config, _: f32, _: *mut *mut tof_error);
    fn tof_camera_config_get_sync_mode(_: *const tof_camera_config, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_set_sync_mode(_: *mut tof_camera_config, _: i32, _: *mut *mut tof_error);
    fn tof_camera_config_maximum_integration_time(_: *mut tof_camera_config, _: usize, _: *const u32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_get_dac(_: *const tof_camera_config, _: *const u16, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_set_dac(_: *mut tof_camera_config, _: *const u16, _: usize, _: *mut *mut tof_error);
    fn tof_camera_config_get_mlx_mipi_speed(_: *const tof_camera_config, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_set_mlx_mipi_speed(_: *mut tof_camera_config, _: i32, _: *mut *mut tof_error);
    fn tof_camera_config_get_mipi_speed(_: *const tof_camera_config, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_set_mipi_speed(_: *mut tof_camera_config, _: i32, _: *mut *mut tof_error);
    fn tof_camera_config_get_mlx_preheat(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> u32;
    fn tof_camera_config_set_mlx_preheat(_: *mut tof_camera_config, _: usize, _: u32, _: *mut *mut tof_error);
    fn tof_camera_config_get_mlx_preheat_enable(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> bool;
    fn tof_camera_config_set_mlx_preheat_enable(_: *mut tof_camera_config, _: usize, _: bool, _: *mut *mut tof_error);
    fn tof_camera_config_set_opt_kea_nios_hex(_: *mut tof_camera_config, _: *const c_char, _: *mut *mut tof_error);
    fn tof_camera_config_get_opt_kea_sub_quad(_: *const tof_camera_config, _: *mut *mut tof_error) -> *const i32;
    fn tof_camera_config_set_opt_kea_sub_quad(_: *mut tof_camera_config, _: *const i32, _: *mut *mut tof_error);
    fn tof_camera_config_get_processing(_: *const tof_camera_config, _: *mut *mut tof_error) -> *mut tof_processing_config;
    fn tof_camera_config_set_processing(_: *mut tof_camera_config, _: *mut tof_processing_config, _: *mut *mut tof_error);
    fn tof_camera_config_get_frame_time(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> u32;
    fn tof_camera_config_set_frame_time(_: *mut tof_camera_config, _: usize, _: u32, _: *mut *mut tof_error);
    fn tof_camera_config_get_global_phase_offset(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> *const f32;
    fn tof_camera_config_set_global_phase_offset(_: *mut tof_camera_config, _: usize, _: *const f32, _: *mut *mut tof_error);
    fn tof_camera_config_get_illumination_mode(_: *const tof_camera_config, _: usize, _: *mut *mut tof_error) -> u8;
    fn tof_camera_config_set_illumination_mode(_: *mut tof_camera_config, _: usize, _: u8, _: *mut *mut tof_error);
    fn tof_camera_config_get_image_offsets(_: *const tof_camera_config, _: usize, _: *const f32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_set_image_offsets(_: *mut tof_camera_config, _: usize, _: *const f32, _: usize, _: *mut *mut tof_error);
    fn tof_camera_config_get_rgb_width(_: *mut tof_camera_config, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_get_rgb_height(_: *mut tof_camera_config, _: *mut *mut tof_error) -> usize;
    fn tof_camera_config_get_trigger_mode(_: *const tof_camera_config, _: *mut *mut tof_error) -> i32;
    fn tof_camera_config_set_trigger_mode(_: *mut tof_camera_config, _: i32, _: *mut *mut tof_error);
    fn tof_camera_config_get_depth_fps(_: *mut tof_camera_config, _: *mut *mut tof_error) -> f32;
}

/// This class allows you to view/edit the camera settings
pub struct CameraConfig {
    ptr: *mut tof_camera_config,
}

impl Drop for CameraConfig {
    fn drop(&mut self) {
        unsafe { tof_camera_config_delete(self.ptr); }
    }
}

unsafe impl Send for CameraConfig {}

impl CameraConfig {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_camera_config {
        self.ptr
    }

    /// Read camera config from disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location of camera config on disk
    ///
    pub fn new_from_disk(file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_camera_config_new_from_disk(file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

    /// Write camera config to disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location to save camera config to
    ///
    pub fn write(&self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_camera_config_write(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the camera type
    ///
    /// ### Returns
    ///
    /// Camera type
    ///
    pub fn get_type(&self) -> Result<CameraType> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_type(self.ptr, err.ptr()) };
        let out: CameraType = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Reset camera config to default
    ///
    pub fn reset(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_reset(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Get the amount of frames in the config. Having multiple frames allows
    /// you to configure different modulation frequencies or integration times
    /// per frame.
    ///
    /// ### Returns
    ///
    /// Frame size
    ///
    pub fn frame_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_frame_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Add a default frame to the camera config
    ///
    pub fn add_frame(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_add_frame(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Generate default processing config from camera configuration
    ///
    /// ### Returns
    ///
    /// The default processing config
    ///
    pub fn default_processing(&self) -> Result<ProcessingConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_default_processing(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(ProcessingConfig { ptr: ret })
        }
    }

    /// Erase specified frame
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    pub fn erase_frame(&mut self, frame: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_erase_frame(self.ptr, frame, err.ptr()) };
        err.empty_result()
    }

    /// Get the phase shifts. Value ranges from 0.0 to 1.0 which translates to
    /// 0-2 pi
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Phase shifts
    ///
    pub fn get_phase_shifts(&self, frame: usize) -> Result<Vec<f32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_config_get_phase_shifts(self.ptr, frame, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<f32> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_config_get_phase_shifts(self.ptr, frame, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the phase shifts. Value ranges from 0.0 to 1.0 which translates to
    /// 0-2 pi
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `phase_shifts` Phase shifts
    ///
    pub fn set_phase_shifts(&mut self, frame: usize, phase_shifts: &[f32]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_phase_shifts(self.ptr, frame, phase_shifts.as_ptr(), phase_shifts.len(), err.ptr()) };
        err.empty_result()
    }

    /// Get the integration time
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Integration time in micro seconds
    ///
    pub fn get_integration_time(&self, frame: usize) -> Result<Vec<u32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_config_get_integration_time(self.ptr, frame, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<u32> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_config_get_integration_time(self.ptr, frame, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the integration time
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `integration_time` Integration time in micro seconds
    ///
    pub fn set_integration_time(&mut self, frame: usize, integration_time: &[u32]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_integration_time(self.ptr, frame, integration_time.as_ptr(), integration_time.len(), err.ptr()) };
        err.empty_result()
    }

    /// Get the mode to set the camera's depth sensor to
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Sensor mode
    ///
    pub fn get_sensor_mode(&self, frame: usize) -> Result<SensorMode> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_sensor_mode(self.ptr, frame, err.ptr()) };
        let out: SensorMode = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the mode to set the camera's depth sensor to
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `sensor_mode` Sensor mode
    ///
    pub fn set_sensor_mode(&mut self, frame: usize, sensor_mode: SensorMode) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_sensor_mode(self.ptr, frame, sensor_mode as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get the modulation frequency
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Modulation frequency in MHz
    ///
    pub fn get_modulation_frequency(&self, frame: usize) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_modulation_frequency(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the modulation frequency
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `modulation_frequency` Modulation frequency in MHz
    ///
    pub fn set_modulation_frequency(&mut self, frame: usize, modulation_frequency: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_modulation_frequency(self.ptr, frame, modulation_frequency, err.ptr()) };
        err.empty_result()
    }

    /// Get the duty cycle
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Duty cycle 0.0 to 1.0
    ///
    pub fn get_duty_cycle(&self, frame: usize) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_duty_cycle(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the duty cycle
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `duty_cycle` Duty cycle 0.0 to 1.0
    ///
    pub fn set_duty_cycle(&mut self, frame: usize, duty_cycle: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_duty_cycle(self.ptr, frame, duty_cycle, err.ptr()) };
        err.empty_result()
    }

    /// Get the region of interest
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Region of interest
    ///
    pub fn get_roi(&self, frame: usize) -> Result<Roi> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_roi(self.ptr, frame, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Roi { ptr: ret })
        }
    }

    /// Set the region of interest
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `roi` Region of interest
    ///
    pub fn set_roi(&mut self, frame: usize, roi: &Roi) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_roi(self.ptr, frame, roi.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the amount of binning
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Amount of binning 0=none, 1=x2, 2=x4, 3=x8
    ///
    pub fn get_binning(&self, frame: usize) -> Result<u8> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_binning(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the amount of binning
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `binning` Amount of binning 0=none, 1=x2, 2=x4, 3=x8
    ///
    pub fn set_binning(&mut self, frame: usize, binning: u8) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_binning(self.ptr, frame, binning, err.ptr()) };
        err.empty_result()
    }

    /// Get is the image flipped
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Image is flipped
    ///
    pub fn get_flip(&self, frame: usize) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_flip(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set is the image flipped
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `flip` Image is flipped
    ///
    pub fn set_flip(&mut self, frame: usize, flip: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_flip(self.ptr, frame, flip, err.ptr()) };
        err.empty_result()
    }

    /// Get is the image mirrored
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Image is mirrored
    ///
    pub fn get_mirror(&self, frame: usize) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_mirror(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set is the image mirrored
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `mirror` Image is mirrored
    ///
    pub fn set_mirror(&mut self, frame: usize, mirror: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_mirror(self.ptr, frame, mirror, err.ptr()) };
        err.empty_result()
    }

    /// Get the image sensor gain
    ///
    /// ### Returns
    ///
    /// Image sensor gain
    ///
    pub fn get_gain(&self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_gain(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the image sensor gain
    ///
    /// ### Arguments
    ///
    /// - `gain` Image sensor gain
    ///
    pub fn set_gain(&mut self, gain: f32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_gain(self.ptr, gain, err.ptr()) };
        err.empty_result()
    }

    /// Get the kind of camera synchronization
    ///
    /// ### Returns
    ///
    /// The kind of sync mode. 0 waits for the vsync pulse before starting the
    /// next frame. 1 optional vsync 2 is no frame syncing
    ///
    pub fn get_sync_mode(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_sync_mode(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the kind of camera synchronization
    ///
    /// ### Arguments
    ///
    /// - `sync_mode` The kind of sync mode. 0 waits for the vsync pulse before
    ///
    pub fn set_sync_mode(&mut self, sync_mode: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_sync_mode(self.ptr, sync_mode, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum integration time possible
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Maximum integration time in micro seconds
    ///
    pub fn maximum_integration_time(&mut self, frame: usize) -> Result<Vec<u32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_config_maximum_integration_time(self.ptr, frame, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<u32> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_config_maximum_integration_time(self.ptr, frame, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the DAC gain
    ///
    /// ### Returns
    ///
    /// The DAC gain
    ///
    pub fn get_dac(&self) -> Result<Vec<u16>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_config_get_dac(self.ptr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<u16> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_config_get_dac(self.ptr, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the DAC gain
    ///
    /// ### Arguments
    ///
    /// - `dac` The DAC gain
    ///
    pub fn set_dac(&mut self, dac: &[u16]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_dac(self.ptr, dac.as_ptr(), dac.len(), err.ptr()) };
        err.empty_result()
    }

    /// Get the mipi speed for the MLX75027
    ///
    /// ### Returns
    ///
    /// The mipi speed. Can only be the values 300, 600, 704, 800 and 960
    ///
    pub fn get_mlx_mipi_speed(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_mlx_mipi_speed(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the mipi speed for the MLX75027
    ///
    /// ### Arguments
    ///
    /// - `mipi_speed` The mipi speed. Can only be the values 300, 600, 704, 800
    ///
    pub fn set_mlx_mipi_speed(&mut self, mipi_speed: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_mlx_mipi_speed(self.ptr, mipi_speed, err.ptr()) };
        err.empty_result()
    }

    /// Get the mipi speed
    ///
    /// ### Returns
    ///
    /// The mipi speed
    ///
    pub fn get_mipi_speed(&self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_mipi_speed(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the mipi speed
    ///
    /// ### Arguments
    ///
    /// - `mipi_speed` The mipi speed
    ///
    pub fn set_mipi_speed(&mut self, mipi_speed: i32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_mipi_speed(self.ptr, mipi_speed, err.ptr()) };
        err.empty_result()
    }

    /// Get the preheat time
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// The preheat time
    ///
    pub fn get_mlx_preheat(&self, frame: usize) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_mlx_preheat(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the preheat time
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `preheat` The preheat time
    ///
    pub fn set_mlx_preheat(&mut self, frame: usize, preheat: u32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_mlx_preheat(self.ptr, frame, preheat, err.ptr()) };
        err.empty_result()
    }

    /// Get enable preheat
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Enable preheating
    ///
    pub fn get_mlx_preheat_enable(&self, frame: usize) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_mlx_preheat_enable(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set enable preheat
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `enable` Enable preheating
    ///
    pub fn set_mlx_preheat_enable(&mut self, frame: usize, enable: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_mlx_preheat_enable(self.ptr, frame, enable, err.ptr()) };
        err.empty_result()
    }

    /// Set the Nios hex for the keaB cameras
    ///
    /// ### Arguments
    ///
    /// - `file_location` File location of the nios hex file
    ///
    pub fn set_opt_kea_nios_hex(&mut self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_camera_config_set_opt_kea_nios_hex(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the opt8241 number of sub and quad frames
    ///
    /// ### Returns
    ///
    /// Number of subs and quads respectively
    ///
    pub fn get_opt_kea_sub_quad(&self) -> Result<[i32; 2]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_opt_kea_sub_quad(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[i32; 2]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 2);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Set the opt8241 number of sub and quad frames
    ///
    /// ### Arguments
    ///
    /// - `subs_quads` Number of subs and quads respectively
    ///
    pub fn set_opt_kea_sub_quad(&mut self, subs_quads: [i32; 2]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_opt_kea_sub_quad(self.ptr, subs_quads.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the processing configuration for when the depth is calculated on
    /// camera
    ///
    /// ### Returns
    ///
    /// The pipeline processing to be done on camera
    ///
    pub fn get_processing(&self) -> Result<ProcessingConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_processing(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(ProcessingConfig { ptr: ret })
        }
    }

    /// Set the processing configuration for when the depth is calculated on
    /// camera
    ///
    /// ### Arguments
    ///
    /// - `processing` The pipeline processing to be done on camera
    ///
    pub fn set_processing(&mut self, processing: &ProcessingConfig) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_processing(self.ptr, processing.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the time this depth frame will take. This is useful when you want a
    /// steady but lower framerate.
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Frame time in micro seconds
    ///
    pub fn get_frame_time(&self, frame: usize) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_frame_time(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the time this depth frame will take. This is useful when you want a
    /// steady but lower framerate.
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `frame_time` Frame time in micro seconds
    ///
    pub fn set_frame_time(&mut self, frame: usize, frame_time: u32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_frame_time(self.ptr, frame, frame_time, err.ptr()) };
        err.empty_result()
    }

    /// Get phase offset for illumination and sensor. The phase offsets are two
    /// floating point values corresponding to the phase delay for the
    /// illumination and sensor respectively. The values should be between 0.0
    /// and 1.0 where 0.0 is no phase offset and 1.0 is 360 degree phase offset
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Image is mirrored
    ///
    pub fn get_global_phase_offset(&self, frame: usize) -> Result<[f32; 2]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_global_phase_offset(self.ptr, frame, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f32; 2]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 2);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Set phase offset for illumination and sensor. The phase offsets are two
    /// floating point values corresponding to the phase delay for the
    /// illumination and sensor respectively. The values should be between 0.0
    /// and 1.0 where 0.0 is no phase offset and 1.0 is 360 degree phase offset
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `phase_offset` Image is mirrored
    ///
    pub fn set_global_phase_offset(&mut self, frame: usize, phase_offset: [f32; 2]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_global_phase_offset(self.ptr, frame, phase_offset.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the illumination mode
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// Illumination mode. 0 is actively modulating, 2 is active low
    /// (illumination disabled) and 3 is active high
    ///
    pub fn get_illumination_mode(&self, frame: usize) -> Result<u8> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_illumination_mode(self.ptr, frame, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the illumination mode
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `illumination_mode` Illumination mode. 0 is actively modulating, 2 is
    ///
    pub fn set_illumination_mode(&mut self, frame: usize, illumination_mode: u8) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_illumination_mode(self.ptr, frame, illumination_mode, err.ptr()) };
        err.empty_result()
    }

    /// Get the image offsets. This configuration is only for sensors with
    /// column phase offsets.
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    ///
    /// ### Returns
    ///
    /// The image offset per column in 0.0 to 1.0
    ///
    pub fn get_image_offsets(&self, frame: usize) -> Result<Vec<f32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_config_get_image_offsets(self.ptr, frame, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<f32> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_config_get_image_offsets(self.ptr, frame, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the image offsets. This configuration is only for sensors with
    /// column phase offsets.
    ///
    /// ### Arguments
    ///
    /// - `frame` Frame number
    /// - `image_offsets` The image offset per column in 0.0 to 1.0
    ///
    pub fn set_image_offsets(&mut self, frame: usize, image_offsets: &[f32]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_image_offsets(self.ptr, frame, image_offsets.as_ptr(), image_offsets.len(), err.ptr()) };
        err.empty_result()
    }

    /// Get the width of the rgb sensor
    ///
    /// ### Returns
    ///
    /// Width in pixels
    ///
    pub fn get_rgb_width(&mut self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_rgb_width(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the height of the rgb sensor
    ///
    /// ### Returns
    ///
    /// Height in pixels
    ///
    pub fn get_rgb_height(&mut self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_rgb_height(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the trigger mode
    ///
    /// ### Returns
    ///
    /// The trigger mode of the camera
    ///
    pub fn get_trigger_mode(&self) -> Result<TriggerMode> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_trigger_mode(self.ptr, err.ptr()) };
        let out: TriggerMode = unsafe { std::mem::transmute(ret) };
        err.make_result(out)
    }

    /// Set the trigger mode
    ///
    /// ### Arguments
    ///
    /// - `trigger_mode` The trigger mode of the camera
    ///
    pub fn set_trigger_mode(&mut self, trigger_mode: TriggerMode) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_config_set_trigger_mode(self.ptr, trigger_mode as i32, err.ptr()) };
        err.empty_result()
    }

    /// Get the maximum theoretical fps the camera can run
    ///
    /// ### Returns
    ///
    /// The fps in Hz
    ///
    pub fn get_depth_fps(&mut self) -> Result<f32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_get_depth_fps(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_camera_config_equal(_: *const tof_camera_config, _: *const tof_camera_config, _: *mut *mut tof_error) -> bool;
}

/// Test whether camera configurations are equal
///
/// ### Arguments
///
/// - `lhs` First camera config
/// - `rhs` Second camera config
///
/// ### Returns
///
/// Camera configurations are equal
///
pub fn camera_config_equal(lhs: &CameraConfig, rhs: &CameraConfig) -> Result<bool> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_camera_config_equal(lhs.ptr(), rhs.ptr(), err.ptr()) };
    err.make_result(ret)
}

#[repr(C)]
struct tof_point_2d {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_point_2d_delete(_: *mut tof_point_2d);
    fn tof_point_2d_row(_: *mut tof_point_2d, _: *mut *mut tof_error) -> i32;
    fn tof_point_2d_col(_: *mut tof_point_2d, _: *mut *mut tof_error) -> i32;
}

/// 2D point for indexing into array
pub struct Point2d {
    ptr: *mut tof_point_2d,
}

impl Drop for Point2d {
    fn drop(&mut self) {
        unsafe { tof_point_2d_delete(self.ptr); }
    }
}

unsafe impl Send for Point2d {}

impl Point2d {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_point_2d {
        self.ptr
    }

    /// Get the row value
    ///
    /// ### Returns
    ///
    /// Row value
    ///
    pub fn row(&mut self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_point_2d_row(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the column value
    ///
    /// ### Returns
    ///
    /// Column value
    ///
    pub fn col(&mut self) -> Result<i32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_point_2d_col(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

}

#[repr(C)]
struct tof_calibration {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_calibration_delete(_: *mut tof_calibration);
    fn tof_calibration_new_from_disk(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_calibration;
    fn tof_calibration_write(_: *const tof_calibration, _: *const c_char, _: *mut *mut tof_error);
    fn tof_calibration_get_calibrated_frequencies(_: *const tof_calibration, _: *const f32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_calibration_get_focal_length_x(_: *const tof_calibration, _: *mut *mut tof_error) -> f64;
    fn tof_calibration_get_focal_length_y(_: *const tof_calibration, _: *mut *mut tof_error) -> f64;
    fn tof_calibration_get_principal_point_x(_: *const tof_calibration, _: *mut *mut tof_error) -> f64;
    fn tof_calibration_get_principal_point_y(_: *const tof_calibration, _: *mut *mut tof_error) -> f64;
    fn tof_calibration_get_depth_camera_matrix(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
    fn tof_calibration_get_depth_distortion_coefficients(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
    fn tof_calibration_get_rgb_camera_matrix(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
    fn tof_calibration_get_rgb_distortion_coefficients(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
    fn tof_calibration_get_rotation_vector(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
    fn tof_calibration_get_translation_vector(_: *mut tof_calibration, _: *mut *mut tof_error) -> *const f64;
}

/// This class contains all the calibration information
pub struct Calibration {
    ptr: *mut tof_calibration,
}

impl Drop for Calibration {
    fn drop(&mut self) {
        unsafe { tof_calibration_delete(self.ptr); }
    }
}

unsafe impl Send for Calibration {}

impl Calibration {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_calibration {
        self.ptr
    }

    /// Read calibration from disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location of calibration file
    ///
    pub fn new_from_disk(file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_calibration_new_from_disk(file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Calibration { ptr: ret })
        }
    }

    /// Write calibration to disk
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location to save the calibration file to
    ///
    pub fn write(&self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_calibration_write(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// The the calibrated frequencies in the calibration file
    ///
    /// ### Returns
    ///
    /// The calibrated modulation frequencies
    ///
    pub fn get_calibrated_frequencies(&self) -> Result<Vec<f32>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_calibration_get_calibrated_frequencies(self.ptr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<f32> = Vec::with_capacity(size);
        let ret = unsafe { tof_calibration_get_calibrated_frequencies(self.ptr, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get focal length x
    ///
    /// ### Returns
    ///
    /// Focal length
    ///
    pub fn get_focal_length_x(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_focal_length_x(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get focal length y
    ///
    /// ### Returns
    ///
    /// Focal length
    ///
    pub fn get_focal_length_y(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_focal_length_y(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get principal point x
    ///
    /// ### Returns
    ///
    /// Principal point
    ///
    pub fn get_principal_point_x(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_principal_point_x(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get principal point y
    ///
    /// ### Returns
    ///
    /// Principal point
    ///
    pub fn get_principal_point_y(&self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_principal_point_y(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the camera matrix of the depth sensor and lens
    ///
    /// ### Returns
    ///
    /// Camera matrix, 3x3 matrix
    ///
    pub fn get_depth_camera_matrix(&mut self) -> Result<[f64; 9]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_depth_camera_matrix(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 9]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 9);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Get the distortion coefficients of the depth lens
    ///
    /// ### Returns
    ///
    /// Distortion coefficients, 5x1 matrix
    ///
    pub fn get_depth_distortion_coefficients(&mut self) -> Result<[f64; 5]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_depth_distortion_coefficients(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 5]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 5);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Get the camera matrix of the rgb sensor and lens
    ///
    /// ### Returns
    ///
    /// Camera matrix, 3x3 matrix
    ///
    pub fn get_rgb_camera_matrix(&mut self) -> Result<[f64; 9]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_rgb_camera_matrix(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 9]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 9);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Get the distortion coefficients of the rgb lens
    ///
    /// ### Returns
    ///
    /// Distortion coefficients, 5x1 matrix
    ///
    pub fn get_rgb_distortion_coefficients(&mut self) -> Result<[f64; 5]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_rgb_distortion_coefficients(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 5]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 5);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Get the rotation vector between the depth and rgb sensor
    ///
    /// ### Returns
    ///
    /// Rotation vector, 3x1 matrix
    ///
    pub fn get_rotation_vector(&mut self) -> Result<[f64; 3]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_rotation_vector(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 3]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 3);
            arr.assume_init()
        };
        err.make_result(arr)
    }

    /// Get the translation vector between the depth and rgb sensor
    ///
    /// ### Returns
    ///
    /// Translation vector, 3x1 matrix
    ///
    pub fn get_translation_vector(&mut self) -> Result<[f64; 3]> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_calibration_get_translation_vector(self.ptr, err.ptr()) };
        let arr = unsafe {
            let mut arr = std::mem::MaybeUninit::<[f64; 3]>::uninit();
            std::ptr::copy_nonoverlapping(ret, (*arr.as_mut_ptr()).as_mut_ptr(), 3);
            arr.assume_init()
        };
        err.make_result(arr)
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_transform_3d_to_depth_2d(_: f32, _: f32, _: f32, _: *mut tof_calibration, _: *mut *mut tof_error) -> *mut tof_point_2d;
    fn tof_transform_3d_to_color_2d(_: f32, _: f32, _: f32, _: *mut tof_calibration, _: *mut *mut tof_error) -> *mut tof_point_2d;
    fn tof_transform_color_2d_to_depth_2d(_: i32, _: i32, _: *mut tof_data, _: *mut tof_calibration, _: *mut *mut tof_error) -> *mut tof_point_2d;
}

/// Transform a 3D point into a 2D pixel coordinate of the depth camera
///
/// ### Arguments
///
/// - `x` X value
/// - `y` Y value
/// - `z` Z value
/// - `calibration` The camera calibration
///
/// ### Returns
///
/// 2D pixel coordinate of the depth camera
///
pub fn transform_3d_to_depth_2d(x: f32, y: f32, z: f32, calibration: &Calibration) -> Result<Point2d> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_transform_3d_to_depth_2d(x, y, z, calibration.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(Point2d { ptr: ret })
    }
}

/// Transform a 3D point into a 2D pixel coordinate of the color camera
///
/// ### Arguments
///
/// - `x` X value
/// - `y` Y value
/// - `z` Z value
/// - `calibration` The camera calibration
///
/// ### Returns
///
/// 2D pixel coordinate of the color camera
///
pub fn transform_3d_to_color_2d(x: f32, y: f32, z: f32, calibration: &Calibration) -> Result<Point2d> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_transform_3d_to_color_2d(x, y, z, calibration.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(Point2d { ptr: ret })
    }
}

/// Transform a 2D pixel coordinate of the color camera to the 2D pixel
/// coordinate of the depth camera.
///
/// ### Arguments
///
/// - `row` Row coordinate
/// - `col` Column cooridnate
/// - `depth_data` Depth data containing xyz information, XYZ, XYZAmp, XYZBGR
/// - `calibration` The camera calibration
///
/// ### Returns
///
/// 2D pixel coordinate of the depth camera
///
pub fn transform_color_2d_to_depth_2d(row: i32, col: i32, depth_data: &Data, calibration: &Calibration) -> Result<Point2d> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_transform_color_2d_to_depth_2d(row, col, depth_data.ptr(), calibration.ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(Point2d { ptr: ret })
    }
}

#[repr(C)]
struct tof_camera {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_camera_start(_: *mut tof_camera, _: *mut *mut tof_error);
    fn tof_camera_stop(_: *mut tof_camera, _: *mut *mut tof_error);
    fn tof_camera_is_streaming(_: *const tof_camera, _: *mut *mut tof_error) -> bool;
    fn tof_camera_is_connected(_: *const tof_camera, _: *mut *mut tof_error) -> bool;
    fn tof_camera_get_stream_list(_: *const tof_camera, _: *const *mut tof_stream, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_set_stream_list(_: *mut tof_camera, _: *const *mut tof_stream, _: usize, _: *mut *mut tof_error);
    fn tof_camera_is_stream_list_set(_: *const tof_camera, _: *mut *mut tof_error) -> bool;
    fn tof_camera_get_set_stream_list(_: *mut tof_camera, _: *const *mut tof_stream, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_get_process_config(_: *const tof_camera, _: *mut *mut tof_error) -> *mut tof_processing_config;
    fn tof_camera_set_process_config(_: *mut tof_camera, _: *mut tof_processing_config, _: *mut *mut tof_error);
    fn tof_camera_has_frames(_: *const tof_camera, _: *mut *mut tof_error) -> bool;
    fn tof_camera_get_frames(_: *mut tof_camera, _: *const *mut tof_data, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_clear_buffer(_: *mut tof_camera, _: *mut *mut tof_error);
    fn tof_camera_get_buffer_size(_: *const tof_camera, _: *mut *mut tof_error) -> usize;
    fn tof_camera_set_buffer_size(_: *mut tof_camera, _: usize, _: *mut *mut tof_error);
    fn tof_camera_get_camera_config(_: *const tof_camera, _: *mut *mut tof_error) -> *mut tof_camera_config;
    fn tof_camera_get_serial(_: *const tof_camera, _: *mut *mut tof_error) -> *const c_char;
    fn tof_camera_get_calibration(_: *const tof_camera, _: *mut *mut tof_error) -> *mut tof_calibration;
    fn tof_camera_get_user_pointer_capacity(_: *const tof_camera, _: *mut *mut tof_error) -> usize;
    fn tof_camera_set_user_pointer_capacity(_: *mut tof_camera, _: usize, _: *mut *mut tof_error);
    fn tof_camera_add_user_pointer(_: *mut tof_camera, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: i32, _: *mut *mut tof_error);
    fn tof_camera_clear_user_pointers(_: *mut tof_camera, _: *mut *mut tof_error);
    fn tof_camera_config_index(_: *const tof_camera, _: *mut *mut tof_error) -> u32;
    fn tof_camera_switch_config(_: *mut tof_camera, _: u32, _: *mut *mut tof_error);
    fn tof_camera_possible_frame_types(_: *const tof_camera, _: *const i32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_get_output_frame_types(_: *const tof_camera, _: *const i32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_set_output_frame_types(_: *mut tof_camera, _: *const i32, _: usize, _: *mut *mut tof_error);
    fn tof_camera_has_multiple_configurations(_: *const tof_camera, _: *mut *mut tof_error) -> bool;
    fn tof_camera_get_multiple_camera_configurations(_: *const tof_camera, _: *const *mut tof_camera_config, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_get_multiple_processing_configurations(_: *const tof_camera, _: *const *mut tof_processing_config, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_get_multiple_names(_: *const tof_camera, _: *const *const c_char, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_camera_get_multiple_descriptions(_: *const tof_camera, _: *const *const c_char, _: usize, _: *mut *mut tof_error) -> usize;
}

/// The main interface to the depth cameras
pub trait Camera : private::Private {
    unsafe fn camera_ptr(&self) -> *mut c_void;

    /// Start streaming of the camera
    ///
    fn start(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_start(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.empty_result()
    }

    /// Stop streaming of the camera
    ///
    fn stop(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_stop(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.empty_result()
    }

    /// Check whether the camera is streaming
    ///
    /// ### Returns
    ///
    /// Is streaming
    ///
    fn is_streaming(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_is_streaming(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Check whether the camera is still connected
    ///
    /// ### Returns
    ///
    /// Is connected
    ///
    fn is_connected(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_is_connected(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the different output streams the camera can provide
    ///
    /// ### Returns
    ///
    /// The streams
    ///
    fn get_stream_list(&self) -> Result<Vec<Stream>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_stream_list(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<Stream> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_stream_list(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the streams that you want to get from the camera
    ///
    /// ### Arguments
    ///
    /// - `streams` The wanted streams
    ///
    fn set_stream_list(&mut self, streams: &[Stream]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_set_stream_list(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*streams.as_ptr()).ptr), streams.len(), err.ptr()) };
        err.empty_result()
    }

    /// Check whether the stream list is set
    ///
    /// ### Returns
    ///
    /// Is stream list set
    ///
    fn is_stream_list_set(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_is_stream_list_set(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the stream list that is set on the camera
    ///
    /// ### Returns
    ///
    /// The currently set streams
    ///
    fn get_set_stream_list(&mut self) -> Result<Vec<Stream>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_set_stream_list(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<Stream> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_set_stream_list(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the processing config of the camera
    ///
    /// ### Returns
    ///
    /// The processing config
    ///
    fn get_process_config(&self) -> Result<ProcessingConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_process_config(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(ProcessingConfig { ptr: ret })
        }
    }

    /// Set the processing config of the camera
    ///
    /// ### Arguments
    ///
    /// - `config` The processing config
    ///
    fn set_process_config(&mut self, config: &ProcessingConfig) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_set_process_config(self.camera_ptr() as *mut tof_camera, config.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Check if the camera has frames available
    ///
    /// ### Returns
    ///
    /// Has frames available
    ///
    fn has_frames(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_has_frames(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Get data frames from camera
    ///
    /// ### Returns
    ///
    /// Data frames
    ///
    fn get_frames(&mut self) -> Result<Vec<Data>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_frames(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<Data> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_frames(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Clear the circular buffer used to store frames
    ///
    fn clear_buffer(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_clear_buffer(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.empty_result()
    }

    /// Get the circular buffer size
    ///
    /// ### Returns
    ///
    /// The maximum number of sets of frames that is stored before the oldest is
    /// overwritten
    ///
    fn get_buffer_size(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_buffer_size(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the circular buffer size
    ///
    /// ### Arguments
    ///
    /// - `size` The maximum number of sets of frames that is stored before the
    ///
    fn set_buffer_size(&mut self, size: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_set_buffer_size(self.camera_ptr() as *mut tof_camera, size, err.ptr()) };
        err.empty_result()
    }

    /// Get the currently active camera config
    ///
    /// ### Returns
    ///
    /// The camera config
    ///
    fn get_camera_config(&self) -> Result<CameraConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_camera_config(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

    /// Get the serial number of the current camera
    ///
    /// ### Returns
    ///
    /// The serial
    ///
    fn get_serial(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_serial(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Get calibration from the camera
    ///
    /// ### Returns
    ///
    /// The calibration
    ///
    fn get_calibration(&self) -> Result<Calibration> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_calibration(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Calibration { ptr: ret })
        }
    }

    /// Get the amount of user pointers that can be stored
    ///
    /// ### Returns
    ///
    /// The capacity
    ///
    fn get_user_pointer_capacity(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_get_user_pointer_capacity(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the amount of user pointers that can be stored
    ///
    /// ### Arguments
    ///
    /// - `capacity` The capacity
    ///
    fn set_user_pointer_capacity(&mut self, capacity: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_set_user_pointer_capacity(self.camera_ptr() as *mut tof_camera, capacity, err.ptr()) };
        err.empty_result()
    }

    /// Add a pointer that will be filled with the specified data type. This
    /// method allows you to fill data straight into your own data type without
    /// any additional copying.
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    /// - `frame_type` The frame type that will be filled into the pointer
    ///
    unsafe fn add_user_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void, frame_type: FrameType) -> Result<()> {
        let mut err = ErrorHelper::new();
        tof_camera_add_user_pointer(self.camera_ptr() as *mut tof_camera, pointer, capacity, callback, callback_user_ptr, frame_type as i32, err.ptr());
        err.empty_result()
    }

    /// Clear all user pointers in memory
    ///
    fn clear_user_pointers(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_clear_user_pointers(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.empty_result()
    }

    /// Get the currently set configuration index
    ///
    /// ### Returns
    ///
    /// config index
    ///
    fn config_index(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_config_index(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Switch to a different config in real time, depending on configuration
    /// and camera, this can take 100-300ms
    ///
    /// ### Arguments
    ///
    /// - `config_index` Config index to switch to
    ///
    fn switch_config(&mut self, config_index: u32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_switch_config(self.camera_ptr() as *mut tof_camera, config_index, err.ptr()) };
        err.empty_result()
    }

    /// Get the frame types that this camera can output
    ///
    /// ### Returns
    ///
    /// Possible frame types
    ///
    fn possible_frame_types(&self) -> Result<Vec<FrameType>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_possible_frame_types(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<FrameType> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_possible_frame_types(self.camera_ptr() as *mut tof_camera, out.as_mut_ptr() as *mut i32, out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the frame types the camera should output
    ///
    /// ### Returns
    ///
    /// Output frame types
    ///
    fn get_output_frame_types(&self) -> Result<Vec<FrameType>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_output_frame_types(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<FrameType> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_output_frame_types(self.camera_ptr() as *mut tof_camera, out.as_mut_ptr() as *mut i32, out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Set the frame types the camera should output
    ///
    /// ### Arguments
    ///
    /// - `output_frame_types` Output frame types
    ///
    fn set_output_frame_types(&mut self, output_frame_types: &[FrameType]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_camera_set_output_frame_types(self.camera_ptr() as *mut tof_camera, output_frame_types.as_ptr() as *const i32, output_frame_types.len(), err.ptr()) };
        err.empty_result()
    }

    /// Check if multiple configurations are set. Switch config can only be used
    /// when this is the case.
    ///
    /// ### Returns
    ///
    /// Is set
    ///
    fn has_multiple_configurations(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_camera_has_multiple_configurations(self.camera_ptr() as *mut tof_camera, err.ptr()) };
        err.make_result(ret)
    }

    /// Get all the set camera configurations
    ///
    /// ### Returns
    ///
    /// Multiple camera configurations
    ///
    fn get_multiple_camera_configurations(&self) -> Result<Vec<CameraConfig>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_multiple_camera_configurations(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<CameraConfig> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_multiple_camera_configurations(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get all the set processing configurations
    ///
    /// ### Returns
    ///
    /// Multiple processing configurations
    ///
    fn get_multiple_processing_configurations(&self) -> Result<Vec<ProcessingConfig>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_multiple_processing_configurations(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<ProcessingConfig> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_multiple_processing_configurations(self.camera_ptr() as *mut tof_camera, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the name for each configuration
    ///
    /// ### Returns
    ///
    /// Names
    ///
    fn get_multiple_names(&self) -> Result<Vec<&str>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_multiple_names(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<*const c_char> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_multiple_names(self.camera_ptr() as *mut tof_camera, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out.into_iter().map(|str| unsafe {CStr::from_ptr(str).to_str().unwrap_or("")}).collect())
    }

    /// Get the description for each configuration
    ///
    /// ### Returns
    ///
    /// Descriptions
    ///
    fn get_multiple_descriptions(&self) -> Result<Vec<&str>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_camera_get_multiple_descriptions(self.camera_ptr() as *mut tof_camera, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<*const c_char> = Vec::with_capacity(size);
        let ret = unsafe { tof_camera_get_multiple_descriptions(self.camera_ptr() as *mut tof_camera, out.as_mut_ptr(), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out.into_iter().map(|str| unsafe {CStr::from_ptr(str).to_str().unwrap_or("")}).collect())
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_select_streams(_: *const tof_camera, _: *const i32, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_select_streams_mod_freq(_: *const tof_camera, _: *const i32, _: usize, _: f32, _: *mut *mut tof_error) -> usize;
}

/// Select the streams you want to get from the camera
///
/// ### Arguments
///
/// - `camera` A camera
/// - `frame_types` The frame types to select
///
/// ### Returns
///
/// The amount of selected streams
///
pub fn select_streams(camera: &impl Camera, frame_types: &[FrameType]) -> Result<usize> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_select_streams(camera.camera_ptr() as *mut tof_camera, frame_types.as_ptr() as *const i32, frame_types.len(), err.ptr()) };
    err.make_result(ret)
}

/// Select the streams you want to get from the camera with the specified
/// modulation frequency
///
/// ### Arguments
///
/// - `camera` A camera
/// - `frame_types` The frame types to select
/// - `mod_freq` The modulation frequency to select
///
/// ### Returns
///
/// The amount of selected streams
///
pub fn select_streams_mod_freq(camera: &impl Camera, frame_types: &[FrameType], mod_freq: f32) -> Result<usize> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_select_streams_mod_freq(camera.camera_ptr() as *mut tof_camera, frame_types.as_ptr() as *const i32, frame_types.len(), mod_freq, err.ptr()) };
    err.make_result(ret)
}

#[repr(C)]
struct tof_kea_camera {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_kea_camera_delete(_: *mut tof_kea_camera);
    fn tof_kea_camera_new(_: *const tof_processing_config, _: *const c_char, _: *mut *mut tof_error) -> *mut tof_kea_camera;
    fn tof_kea_camera_new_simple(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_kea_camera;
    fn tof_kea_camera_new_both_configs(_: *const tof_processing_config, _: *const c_char, _: *const tof_camera_config, _: *mut *mut tof_error) -> *mut tof_kea_camera;
    fn tof_kea_camera_on_camera_processing_capable(_: *const tof_kea_camera, _: *mut *mut tof_error) -> bool;
    fn tof_kea_camera_get_on_camera_processing(_: *const tof_kea_camera, _: *mut *mut tof_error) -> bool;
    fn tof_kea_camera_set_on_camera_processing(_: *mut tof_kea_camera, _: bool, _: *mut *mut tof_error);
    fn tof_kea_camera_get_delay(_: *const tof_kea_camera, _: *mut *mut tof_error) -> u32;
    fn tof_kea_camera_set_delay(_: *mut tof_kea_camera, _: u32, _: *mut *mut tof_error);
    fn tof_kea_camera_get_packet_size(_: *const tof_kea_camera, _: *mut *mut tof_error) -> u16;
    fn tof_kea_camera_set_packet_size(_: *mut tof_kea_camera, _: u16, _: *mut *mut tof_error);
    fn tof_kea_camera_set_camera_config(_: *mut tof_kea_camera, _: *const tof_camera_config, _: *mut *mut tof_error);
    fn tof_kea_camera_set_configurations(_: *mut tof_kea_camera, _: *const tof_camera_config, _: *const tof_processing_config, _: *mut *mut tof_error);
    fn tof_kea_camera_version(_: *mut tof_kea_camera, _: *mut *mut tof_error) -> *const c_char;
    fn tof_kea_camera_software_trigger_capable(_: *mut tof_kea_camera, _: *mut *mut tof_error) -> bool;
    fn tof_kea_camera_software_trigger(_: *mut tof_kea_camera, _: *mut *mut tof_error);
    fn tof_kea_camera_remote_update_capable(_: *mut tof_kea_camera, _: *mut *mut tof_error) -> bool;
    fn tof_kea_camera_remote_update(_: *mut tof_kea_camera, _: *const c_char, _: *mut *mut tof_error);
    fn tof_kea_camera_multiple_configurations_possible(_: *const tof_kea_camera, _: *mut *mut tof_error) -> bool;
    fn tof_kea_camera_set_multiple_configurations(_: *mut tof_kea_camera, _: *const *mut tof_camera_config, _: usize, _: *const *mut tof_processing_config, _: usize, _: *mut *mut tof_error);
}

/// The main interface to the kea camera
pub struct KeaCamera {
    ptr: *mut tof_kea_camera,
}

impl Drop for KeaCamera {
    fn drop(&mut self) {
        unsafe { tof_kea_camera_delete(self.ptr); }
    }
}

unsafe impl Send for KeaCamera {}

impl KeaCamera {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_kea_camera {
        self.ptr
    }

    /// Construct the kea camera
    ///
    /// ### Arguments
    ///
    /// - `processing_config` The processing config
    /// - `serial` The serial number of the camera to connect to, if the camera
    /// is available over both usb and gige it will be connected over usb. If an
    /// empty string is supplied the camera will connect to the first usb
    /// camera, if no usb camera is connected it will connect to the first gige
    /// camera that replies (this reply time is random, so you might connect to
    /// a different camera each time if you've got multiple cameras connected to
    /// your network). If no camera is found with the given serial number a
    ///
    pub fn new(processing_config: &ProcessingConfig, serial: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let serial_str = CString::new(serial)?;
        let ret = unsafe { tof_kea_camera_new(processing_config.ptr(), serial_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(KeaCamera { ptr: ret })
        }
    }

    /// Construct the kea camera with a processing config generated by
    /// default_processing from the camera config
    ///
    /// ### Arguments
    ///
    /// - `serial` The serial number of the camera to connect to, if the camera
    /// is available over both usb and gige it will be connected over usb. If an
    /// empty string is supplied the camera will connect to the first usb
    /// camera, if no usb camera is connected it will connect to the first gige
    /// camera that replies (this reply time is random, so you might connect to
    /// a different camera each time if you've got multiple cameras connected to
    /// your network). If no camera is found with the given serial number a
    ///
    pub fn new_simple(serial: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let serial_str = CString::new(serial)?;
        let ret = unsafe { tof_kea_camera_new_simple(serial_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(KeaCamera { ptr: ret })
        }
    }

    /// Construct the kea camera with the given processing and camera config.
    ///
    /// ### Arguments
    ///
    /// - `processing_config` The processing config
    /// - `serial` The serial number of the camera to connect to, if the camera
    /// is available over both usb and gige it will be connected over usb. If an
    /// empty string is supplied the camera will connect to the first usb
    /// camera, if no usb camera is connected it will connect to the first gige
    /// camera that replies (this reply time is random, so you might connect to
    /// a different camera each time if you've got multiple cameras connected to
    /// your network). If no camera is found with the given serial number a
    /// - `camera_config` The camera config
    ///
    pub fn new_both_configs(processing_config: &ProcessingConfig, serial: &str, camera_config: &CameraConfig) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let serial_str = CString::new(serial)?;
        let ret = unsafe { tof_kea_camera_new_both_configs(processing_config.ptr(), serial_str.as_ptr(), camera_config.ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(KeaCamera { ptr: ret })
        }
    }

    /// Check whether this kea camera can do on camera processing
    ///
    /// ### Returns
    ///
    /// On camera processing capable
    ///
    pub fn on_camera_processing_capable(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_on_camera_processing_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get on camera processing
    ///
    /// ### Returns
    ///
    /// True on camera processing, false on host processing
    ///
    pub fn get_on_camera_processing(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_get_on_camera_processing(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set on camera processing
    ///
    /// ### Arguments
    ///
    /// - `on_camera_processing` True on camera processing, false on host
    ///
    pub fn set_on_camera_processing(&mut self, on_camera_processing: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_on_camera_processing(self.ptr, on_camera_processing, err.ptr()) };
        err.empty_result()
    }

    /// Get delay between network packets
    ///
    /// ### Returns
    ///
    /// The delay, depending on the camera software it is in nanoseconds or in
    /// cpu instruction
    ///
    pub fn get_delay(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_get_delay(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set delay between network packets
    ///
    /// ### Arguments
    ///
    /// - `delay` The delay, depending on the camera software it is in
    ///
    pub fn set_delay(&mut self, delay: u32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_delay(self.ptr, delay, err.ptr()) };
        err.empty_result()
    }

    /// Get maximum size of each network packet transmitted
    ///
    /// ### Returns
    ///
    /// Packet size in bytes
    ///
    pub fn get_packet_size(&self) -> Result<u16> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_get_packet_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set maximum size of each network packet transmitted
    ///
    /// ### Arguments
    ///
    /// - `size` Packet size in bytes
    ///
    pub fn set_packet_size(&mut self, size: u16) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_packet_size(self.ptr, size, err.ptr()) };
        err.empty_result()
    }

    /// Set the camera config
    ///
    /// ### Arguments
    ///
    /// - `camera_config` Camera config
    ///
    pub fn set_camera_config(&mut self, camera_config: &CameraConfig) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_camera_config(self.ptr, camera_config.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Set both the camera and processing config
    ///
    /// ### Arguments
    ///
    /// - `camera_config` Camera config
    /// - `processing_config` Processing config
    ///
    pub fn set_configurations(&mut self, camera_config: &CameraConfig, processing_config: &ProcessingConfig) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_configurations(self.ptr, camera_config.ptr(), processing_config.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the tof library version running on the camera
    ///
    /// ### Returns
    ///
    /// Version
    ///
    pub fn version(&mut self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_version(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Check whether the camera is capable of software trigger
    ///
    /// ### Returns
    ///
    /// Software trigger capable
    ///
    pub fn software_trigger_capable(&mut self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_software_trigger_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Software trigger the camera
    ///
    pub fn software_trigger(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_software_trigger(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Check whether the camera is capable of updating remotely
    ///
    /// ### Returns
    ///
    /// Remote update capable
    ///
    pub fn remote_update_capable(&mut self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_remote_update_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Update the camera remotely. After uploading the update file the function
    /// will return. If a correct update file was uploaded, the camera will
    /// flash the update and reboot. This generally takes about a minute. The
    /// camera can also be updated via a web interface by visiting the ip
    /// address of the camera on port 8080, for example
    /// http://192.168.1.208:8080
    ///
    /// ### Arguments
    ///
    /// - `file_location` Location of the update file
    ///
    pub fn remote_update(&mut self, file_location: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        unsafe { tof_kea_camera_remote_update(self.ptr, file_location_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Check whether camera supports multiple configurations
    ///
    /// ### Returns
    ///
    /// Multiple configurations possible
    ///
    pub fn multiple_configurations_possible(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_kea_camera_multiple_configurations_possible(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set multiple configurations
    ///
    /// ### Arguments
    ///
    /// - `camera_configs` Camera configurations
    /// - `pro_configs` Processing configurations
    ///
    pub fn set_multiple_configurations(&mut self, camera_configs: &[CameraConfig], pro_configs: &[ProcessingConfig]) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_kea_camera_set_multiple_configurations(self.ptr, std::ptr::addr_of!((*camera_configs.as_ptr()).ptr), camera_configs.len(), std::ptr::addr_of!((*pro_configs.as_ptr()).ptr), pro_configs.len(), err.ptr()) };
        err.empty_result()
    }

}

impl Camera for KeaCamera {
    unsafe fn camera_ptr(&self) -> *mut c_void {
        self.ptr as *mut c_void
    }
}

impl private::Private for KeaCamera {}

#[repr(C)]
struct tof_discovered_kea {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_discovered_kea_delete(_: *mut tof_discovered_kea);
    fn tof_discovered_kea_serial(_: *const tof_discovered_kea, _: *mut *mut tof_error) -> *const c_char;
    fn tof_discovered_kea_info(_: *const tof_discovered_kea, _: *mut *mut tof_error) -> *const c_char;
}

/// Simple class containing information about found kea cameras
pub struct DiscoveredKea {
    ptr: *mut tof_discovered_kea,
}

impl Drop for DiscoveredKea {
    fn drop(&mut self) {
        unsafe { tof_discovered_kea_delete(self.ptr); }
    }
}

unsafe impl Send for DiscoveredKea {}

impl DiscoveredKea {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_discovered_kea {
        self.ptr
    }

    /// The serial
    ///
    /// ### Returns
    ///
    /// Serial
    ///
    pub fn serial(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_discovered_kea_serial(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Additional information about the discovered camera
    ///
    /// ### Returns
    ///
    /// Additional information
    ///
    pub fn info(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_discovered_kea_info(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_discover_kea_cameras(_: *const *mut tof_discovered_kea, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_create_kea_camera_gige(_: *const tof_processing_config, _: *const c_char, _: u16, _: *mut *mut tof_error) -> *mut tof_kea_camera;
}

/// Discover all available kea cameras
///
/// ### Returns
///
/// All available kea cameras
///
pub fn discover_kea_cameras() -> Result<Vec<DiscoveredKea>> {
    let mut err = ErrorHelper::new();
    let size = unsafe { tof_discover_kea_cameras(std::ptr::null_mut(), 0, err.ptr()) };
    if err.has_ptr() {
        return Err(err.make_error());
    }
    let mut out: Vec<DiscoveredKea> = Vec::with_capacity(size);
    let ret = unsafe { tof_discover_kea_cameras(std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
    unsafe { out.set_len(ret) };
    err.make_result(out)
}

/// Construct a kea camera using the gige interface. This could be useful in two
/// cases. The first is when the camera is connected with both usb and ethernet
/// but you want to use ethernet. The second is when for some reason the default
/// maximum packet size is too large.
///
/// ### Arguments
///
/// - `processing_config` The processing config
/// - `serial` The serial number of the camera to connect to, if the camera is
/// available over both usb and gige it will be connected over usb. If an empty
/// string is supplied the camera will connect to the first usb camera, if no
/// usb camera is connected it will connect to the first gige camera that
/// replies (this reply time is random, so you might connect to a different
/// camera each time if you've got multiple cameras connected to your network).
/// If no camera is found with the given serial number a nullptr and error is
/// - `packet_size` The packet size to use for communication over ethernet
///
/// ### Returns
///
/// A kea camera
///
pub fn create_kea_camera_gige(processing_config: &ProcessingConfig, serial: &str, opt_packet_size: Option<u16>) -> Result<KeaCamera> {
    let mut err = ErrorHelper::new();
    let serial_str = CString::new(serial)?;
    let packet_size = opt_packet_size.unwrap_or(1472);
    let ret = unsafe { tof_create_kea_camera_gige(processing_config.ptr(), serial_str.as_ptr(), packet_size, err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(KeaCamera { ptr: ret })
    }
}

pub type LogCallback = extern "C" fn(*const c_char, *mut c_void);

pub type LogCallbackSeparated = extern "C" fn(i16, *const c_char, *const c_char, *mut c_void);

#[repr(C)]
struct tof_crash_report {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_crash_report_delete(_: *mut tof_crash_report);
    fn tof_crash_report_new(_: *const c_char, _: *const c_char, _: *mut *mut tof_error) -> *mut tof_crash_report;
    fn tof_crash_report_get_description(_: *const tof_crash_report, _: *mut *mut tof_error) -> *const c_char;
    fn tof_crash_report_set_description(_: *mut tof_crash_report, _: *const c_char, _: *mut *mut tof_error);
    fn tof_crash_report_get_report(_: *const tof_crash_report, _: *mut *mut tof_error) -> *const c_char;
    fn tof_crash_report_save(_: *const tof_crash_report, _: *const c_char, _: *mut *mut tof_error);
}

/// Class containing log information about the crash that occurred
pub struct CrashReport {
    ptr: *mut tof_crash_report,
}

impl Drop for CrashReport {
    fn drop(&mut self) {
        unsafe { tof_crash_report_delete(self.ptr); }
    }
}

unsafe impl Send for CrashReport {}

impl CrashReport {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_crash_report {
        self.ptr
    }

    /// Generate crash report on a camera
    ///
    /// ### Arguments
    ///
    /// - `serial` Camera to generate the crash report on
    /// - `description` Description of what happened
    ///
    pub fn new(serial: &str, description: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let serial_str = CString::new(serial)?;
        let description_str = CString::new(description)?;
        let ret = unsafe { tof_crash_report_new(serial_str.as_ptr(), description_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CrashReport { ptr: ret })
        }
    }

    /// Get the description
    ///
    /// ### Returns
    ///
    /// Description
    ///
    pub fn get_description(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_crash_report_get_description(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Set the description
    ///
    /// ### Arguments
    ///
    /// - `description` Description
    ///
    pub fn set_description(&mut self, description: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let description_str = CString::new(description)?;
        unsafe { tof_crash_report_set_description(self.ptr, description_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the complete report
    ///
    /// ### Returns
    ///
    /// String of the entire report
    ///
    pub fn get_report(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_crash_report_get_report(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Save report to disk
    ///
    /// ### Arguments
    ///
    /// - `file_name` File to save the report to
    ///
    pub fn save(&self, file_name: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let file_name_str = CString::new(file_name)?;
        unsafe { tof_crash_report_save(self.ptr, file_name_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_get_version(_: *mut *mut tof_error) -> *const c_char;
    fn tof_log_to_console(_: *mut *mut tof_error);
    fn tof_log_to_file(_: *const c_char, _: *mut *mut tof_error);
    fn tof_log_callback(_: LogCallback, _: *mut c_void, _: *mut *mut tof_error);
    fn tof_log_callback_separated(_: LogCallbackSeparated, _: *mut c_void, _: *mut *mut tof_error);
    fn tof_log_drop(_: *mut *mut tof_error);
}

/// Get the version of the tof library
///
/// ### Returns
///
/// Library version
///
pub fn get_version() -> Result<&'static str> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_get_version(err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
    Ok(unsafe { CStr::from_ptr(ret).to_str()? })
    }
}

/// Log info/warning/error occuring in the tof library to stdout
///
pub fn log_to_console() -> Result<()> {
    let mut err = ErrorHelper::new();
    unsafe { tof_log_to_console(err.ptr()) };
    err.empty_result()
}

/// Log info/warning/error occuring in the tof library
///
/// ### Arguments
///
/// - `file_location` Location where to store the log file
///
pub fn log_to_file(file_location: &str) -> Result<()> {
    let mut err = ErrorHelper::new();
    let file_location_str = CString::new(file_location)?;
    unsafe { tof_log_to_file(file_location_str.as_ptr(), err.ptr()) };
    err.empty_result()
}

/// Log info/warning/error occuring in the tof library to function
///
/// ### Arguments
///
/// - `callback` Function that is called with log data
///
pub unsafe fn log_callback(callback: LogCallback, callback_user_ptr: *mut c_void) -> Result<()> {
    let mut err = ErrorHelper::new();
    tof_log_callback(callback, callback_user_ptr, err.ptr());
    err.empty_result()
}

/// Log info/warning/error occuring in the tof library to function
///
/// ### Arguments
///
/// - `callback_separated` Function that is called with separated log data
///
pub unsafe fn log_callback_separated(callback_separated: LogCallbackSeparated, callback_separated_user_ptr: *mut c_void) -> Result<()> {
    let mut err = ErrorHelper::new();
    tof_log_callback_separated(callback_separated, callback_separated_user_ptr, err.ptr());
    err.empty_result()
}

/// Stop logging
///
pub fn log_drop() -> Result<()> {
    let mut err = ErrorHelper::new();
    unsafe { tof_log_drop(err.ptr()) };
    err.empty_result()
}

#[repr(C)]
struct tof_csf_camera {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_csf_camera_delete(_: *mut tof_csf_camera);
    fn tof_csf_camera_new(_: *const tof_processing_config, _: *const c_char, _: *mut *mut tof_error) -> *mut tof_csf_camera;
    fn tof_csf_camera_new_simple(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_csf_camera;
    fn tof_csf_camera_get_frames_at(_: *mut tof_csf_camera, _: usize, _: *const *mut tof_data, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_csf_camera_get_current_frame(_: *const tof_csf_camera, _: *mut *mut tof_error) -> usize;
    fn tof_csf_camera_get_frame_number(_: *const tof_csf_camera, _: *mut *mut tof_error) -> usize;
    fn tof_csf_camera_get_fps(_: *const tof_csf_camera, _: *mut *mut tof_error) -> usize;
    fn tof_csf_camera_set_fps(_: *mut tof_csf_camera, _: usize, _: *mut *mut tof_error);
}

/// The csf camera class contains all the logic to generate a stream of depth
/// data from a csf file
pub struct CsfCamera {
    ptr: *mut tof_csf_camera,
}

impl Drop for CsfCamera {
    fn drop(&mut self) {
        unsafe { tof_csf_camera_delete(self.ptr); }
    }
}

unsafe impl Send for CsfCamera {}

impl CsfCamera {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_csf_camera {
        self.ptr
    }

    /// Construct a csf camera from a csf file
    ///
    /// ### Arguments
    ///
    /// - `processing_config` A processing config
    /// - `file_location` File location of the csf file
    ///
    pub fn new(processing_config: &ProcessingConfig, file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_csf_camera_new(processing_config.ptr(), file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CsfCamera { ptr: ret })
        }
    }

    /// Construct a csf camera from a csf file with a processing config
    /// generated by default_processing from the camera config
    ///
    /// ### Arguments
    ///
    /// - `file_location` File location of the csf file
    ///
    pub fn new_simple(file_location: &str) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let file_location_str = CString::new(file_location)?;
        let ret = unsafe { tof_csf_camera_new_simple(file_location_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CsfCamera { ptr: ret })
        }
    }

    /// Get frames at the specified position
    ///
    /// ### Arguments
    ///
    /// - `frame_nr` Which set of frames to process
    ///
    /// ### Returns
    ///
    /// Data frames
    ///
    pub fn get_frames_at(&mut self, frame_nr: usize) -> Result<Vec<Data>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_csf_camera_get_frames_at(self.ptr, frame_nr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<Data> = Vec::with_capacity(size);
        let ret = unsafe { tof_csf_camera_get_frames_at(self.ptr, frame_nr, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Get the location (count) of the current frame
    ///
    /// ### Returns
    ///
    /// The currenc frame location
    ///
    pub fn get_current_frame(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_camera_get_current_frame(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the amount of frame lists this csf file can open/generate
    ///
    /// ### Returns
    ///
    /// The number of frames in the csf file
    ///
    pub fn get_frame_number(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_camera_get_frame_number(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the FPS If the FPS is greater than 0 the time between subsequent
    /// function calls to get_frames or get_frames_at will be limited to the FPS
    /// This is done by sleeping on the next call to get_frames() if the time
    /// between calls is smaller than the time duration dictated by the fps.

    ///
    /// ### Returns
    ///
    /// The frames per second
    ///
    pub fn get_fps(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_csf_camera_get_fps(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the FPS If the FPS is greater than 0 the time between subsequent
    /// function calls to get_frames or get_frames_at will be limited to the FPS
    /// This is done by sleeping on the next call to get_frames() if the time
    /// between calls is smaller than the time duration dictated by the fps.

    ///
    /// ### Arguments
    ///
    /// - `fps` The frames per second
    ///
    pub fn set_fps(&mut self, fps: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_csf_camera_set_fps(self.ptr, fps, err.ptr()) };
        err.empty_result()
    }

}

impl Camera for CsfCamera {
    unsafe fn camera_ptr(&self) -> *mut c_void {
        self.ptr as *mut c_void
    }
}

impl private::Private for CsfCamera {}

/// The possible depth streams that can be streamed from the camera
#[repr(i32)]
#[derive(PartialEq, Debug, Clone, Copy)]
pub enum DepthStreamType {
    /// Stream raw data
    Raw = 0,
    /// Stream both radial and intensity
    RadialIntensity = 1,
    /// Stream radial
    Radial = 2,
    /// Stream intensity
    Intensity = 3,
}

#[repr(C)]
struct tof_discovery_message {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_discovery_message_delete(_: *mut tof_discovery_message);
    fn tof_discovery_message_ip(_: *const tof_discovery_message, _: *mut *mut tof_error) -> *const c_char;
    fn tof_discovery_message_serial(_: *const tof_discovery_message, _: *mut *mut tof_error) -> *const c_char;
}

/// A gige discovery message containing camera information
pub struct DiscoveryMessage {
    ptr: *mut tof_discovery_message,
}

impl Drop for DiscoveryMessage {
    fn drop(&mut self) {
        unsafe { tof_discovery_message_delete(self.ptr); }
    }
}

unsafe impl Send for DiscoveryMessage {}

impl DiscoveryMessage {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_discovery_message {
        self.ptr
    }

    /// Get the camera's ip address
    ///
    /// ### Returns
    ///
    /// Ip address
    ///
    pub fn ip(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_discovery_message_ip(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Get the camera's serial
    ///
    /// ### Returns
    ///
    /// Serial
    ///
    pub fn serial(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_discovery_message_serial(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

}

#[repr(C)]
struct tof_persistent_ip {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_persistent_ip_delete(_: *mut tof_persistent_ip);
    fn tof_persistent_ip_new(_: *const c_char, _: *const c_char, _: *const c_char, _: bool, _: *mut *mut tof_error) -> *mut tof_persistent_ip;
    fn tof_persistent_ip_get_ip(_: *const tof_persistent_ip, _: *mut *mut tof_error) -> *const c_char;
    fn tof_persistent_ip_set_ip(_: *mut tof_persistent_ip, _: *const c_char, _: *mut *mut tof_error);
    fn tof_persistent_ip_get_netmask(_: *const tof_persistent_ip, _: *mut *mut tof_error) -> *const c_char;
    fn tof_persistent_ip_set_netmask(_: *mut tof_persistent_ip, _: *const c_char, _: *mut *mut tof_error);
    fn tof_persistent_ip_get_gateway(_: *const tof_persistent_ip, _: *mut *mut tof_error) -> *const c_char;
    fn tof_persistent_ip_set_gateway(_: *mut tof_persistent_ip, _: *const c_char, _: *mut *mut tof_error);
    fn tof_persistent_ip_get_enabled(_: *const tof_persistent_ip, _: *mut *mut tof_error) -> bool;
    fn tof_persistent_ip_set_enabled(_: *mut tof_persistent_ip, _: bool, _: *mut *mut tof_error);
}

/// A persistent ip data structure
pub struct PersistentIp {
    ptr: *mut tof_persistent_ip,
}

impl Drop for PersistentIp {
    fn drop(&mut self) {
        unsafe { tof_persistent_ip_delete(self.ptr); }
    }
}

unsafe impl Send for PersistentIp {}

impl PersistentIp {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_persistent_ip {
        self.ptr
    }

    /// Create persistent ip
    ///
    /// ### Arguments
    ///
    /// - `ip` IP address
    /// - `netmask` Netmask
    /// - `gateway` Gateway
    /// - `enabled` Enable or disable persistent ip
    ///
    pub fn new(ip: &str, netmask: &str, gateway: &str, enabled: bool) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let ip_str = CString::new(ip)?;
        let netmask_str = CString::new(netmask)?;
        let gateway_str = CString::new(gateway)?;
        let ret = unsafe { tof_persistent_ip_new(ip_str.as_ptr(), netmask_str.as_ptr(), gateway_str.as_ptr(), enabled, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(PersistentIp { ptr: ret })
        }
    }

    /// Get the IP address
    ///
    /// ### Returns
    ///
    /// the IP address
    ///
    pub fn get_ip(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_persistent_ip_get_ip(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Set the IP address
    ///
    /// ### Arguments
    ///
    /// - `ip` the IP address
    ///
    pub fn set_ip(&mut self, ip: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let ip_str = CString::new(ip)?;
        unsafe { tof_persistent_ip_set_ip(self.ptr, ip_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the netmask
    ///
    /// ### Returns
    ///
    /// The netmask
    ///
    pub fn get_netmask(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_persistent_ip_get_netmask(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Set the netmask
    ///
    /// ### Arguments
    ///
    /// - `netmask` The netmask
    ///
    pub fn set_netmask(&mut self, netmask: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let netmask_str = CString::new(netmask)?;
        unsafe { tof_persistent_ip_set_netmask(self.ptr, netmask_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the gateway
    ///
    /// ### Returns
    ///
    /// The gateway
    ///
    pub fn get_gateway(&self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_persistent_ip_get_gateway(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Set the gateway
    ///
    /// ### Arguments
    ///
    /// - `gateway` The gateway
    ///
    pub fn set_gateway(&mut self, gateway: &str) -> Result<()> {
        let mut err = ErrorHelper::new();
        let gateway_str = CString::new(gateway)?;
        unsafe { tof_persistent_ip_set_gateway(self.ptr, gateway_str.as_ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get if persistent ip is enabled
    ///
    /// ### Returns
    ///
    /// Is enabled
    ///
    pub fn get_enabled(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_persistent_ip_get_enabled(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set if persistent ip is enabled
    ///
    /// ### Arguments
    ///
    /// - `enabled` Is enabled
    ///
    pub fn set_enabled(&mut self, enabled: bool) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_persistent_ip_set_enabled(self.ptr, enabled, err.ptr()) };
        err.empty_result()
    }

}

#[repr(C)]
struct tof_gige_interface {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_gige_interface_delete(_: *mut tof_gige_interface);
    fn tof_gige_interface_new(_: u16, _: *mut *mut tof_error) -> *mut tof_gige_interface;
    fn tof_gige_interface_discover(_: *mut tof_gige_interface, _: *const *mut tof_discovery_message, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_gige_interface_discover_one(_: *mut tof_gige_interface, _: *mut *mut tof_error) -> *mut tof_discovery_message;
    fn tof_gige_interface_find(_: *mut tof_gige_interface, _: *const c_char, _: *mut *mut tof_error) -> *mut tof_discovery_message;
    fn tof_gige_interface_connect(_: *mut tof_gige_interface, _: *const tof_discovery_message, _: *mut *mut tof_error);
    fn tof_gige_interface_disconnect(_: *mut tof_gige_interface, _: *mut *mut tof_error);
    fn tof_gige_interface_is_connected(_: *const tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_get_stream_packet_size(_: *const tof_gige_interface, _: *mut *mut tof_error) -> u16;
    fn tof_gige_interface_set_stream_packet_size(_: *mut tof_gige_interface, _: u16, _: *mut *mut tof_error);
    fn tof_gige_interface_test_stream_packet_size(_: *mut tof_gige_interface, _: u16, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_get_delay(_: *const tof_gige_interface, _: *mut *mut tof_error) -> u32;
    fn tof_gige_interface_set_delay(_: *mut tof_gige_interface, _: u32, _: *mut *mut tof_error);
    fn tof_gige_interface_download_configuration(_: *mut tof_gige_interface, _: u16, _: *mut *mut tof_error) -> *mut tof_camera_config;
    fn tof_gige_interface_download_calibration(_: *mut tof_gige_interface, _: u16, _: *mut *mut tof_error) -> *mut tof_calibration;
    fn tof_gige_interface_upload_configuration(_: *mut tof_gige_interface, _: *const tof_camera_config, _: u16, _: *mut *mut tof_error);
    fn tof_gige_interface_depth_stream_capable(_: *const tof_gige_interface, _: i32, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_start_depth_stream(_: *mut tof_gige_interface, _: i32, _: u16, _: *mut *mut tof_error);
    fn tof_gige_interface_stop_depth_stream(_: *mut tof_gige_interface, _: *mut *mut tof_error);
    fn tof_gige_interface_has_depth_data(_: *const tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_get_depth_data(_: *mut tof_gige_interface, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_gige_interface_image_stream_capable(_: *const tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_start_image_stream(_: *mut tof_gige_interface, _: u16, _: *mut *mut tof_error);
    fn tof_gige_interface_stop_image_stream(_: *mut tof_gige_interface, _: *mut *mut tof_error);
    fn tof_gige_interface_has_image_data(_: *const tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_get_image_data(_: *mut tof_gige_interface, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_gige_interface_get_image_data_into_pointer(_: *mut tof_gige_interface, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_gige_interface_persistent_ip_capable(_: *const tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_get_persistent_ip(_: *const tof_gige_interface, _: *mut *mut tof_error) -> *mut tof_persistent_ip;
    fn tof_gige_interface_set_persistent_ip(_: *mut tof_gige_interface, _: *const tof_persistent_ip, _: *mut *mut tof_error);
    fn tof_gige_interface_get_user_pointer_capacity(_: *const tof_gige_interface, _: *mut *mut tof_error) -> usize;
    fn tof_gige_interface_set_user_pointer_capacity(_: *mut tof_gige_interface, _: usize, _: *mut *mut tof_error);
    fn tof_gige_interface_add_depth_user_pointer(_: *mut tof_gige_interface, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: *mut *mut tof_error);
    fn tof_gige_interface_version(_: *mut tof_gige_interface, _: *mut *mut tof_error) -> *const c_char;
    fn tof_gige_interface_software_trigger_capable(_: *mut tof_gige_interface, _: *mut *mut tof_error) -> bool;
    fn tof_gige_interface_software_trigger(_: *mut tof_gige_interface, _: *mut *mut tof_error);
}

/// DEPRECATED, use KeaCamera. Connect to a camera directly using the gige
/// interface
pub struct GigeInterface {
    ptr: *mut tof_gige_interface,
}

impl Drop for GigeInterface {
    fn drop(&mut self) {
        unsafe { tof_gige_interface_delete(self.ptr); }
    }
}

unsafe impl Send for GigeInterface {}

impl GigeInterface {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_gige_interface {
        self.ptr
    }

    /// Create an instance of the gige interface
    ///
    /// ### Arguments
    ///
    /// - `port` The port to use for GVCP communication, if 0 is specified it
    ///
    pub fn new(opt_port: Option<u16>) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        let ret = unsafe { tof_gige_interface_new(port, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(GigeInterface { ptr: ret })
        }
    }

    /// Detect all the cameras on the network
    ///
    /// ### Returns
    ///
    /// A list of discovery messages
    ///
    pub fn discover(&mut self) -> Result<Vec<DiscoveryMessage>> {
        let mut err = ErrorHelper::new();
        let size = unsafe { tof_gige_interface_discover(self.ptr, std::ptr::null_mut(), 0, err.ptr()) };
        if err.has_ptr() {
            return Err(err.make_error());
        }
        let mut out: Vec<DiscoveryMessage> = Vec::with_capacity(size);
        let ret = unsafe { tof_gige_interface_discover(self.ptr, std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
        unsafe { out.set_len(ret) };
        err.make_result(out)
    }

    /// Returns a message of the first camera that responded
    ///
    /// ### Returns
    ///
    /// A discovery message
    ///
    pub fn discover_one(&mut self) -> Result<DiscoveryMessage> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_discover_one(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(DiscoveryMessage { ptr: ret })
        }
    }

    /// Detect a specific camera on the network, will throw an error if it can't
    /// find the camera
    ///
    /// ### Arguments
    ///
    /// - `serial` Serial number of camera to find
    ///
    /// ### Returns
    ///
    /// A discovery message
    ///
    pub fn find(&mut self, serial: &str) -> Result<DiscoveryMessage> {
        let mut err = ErrorHelper::new();
        let serial_str = CString::new(serial)?;
        let ret = unsafe { tof_gige_interface_find(self.ptr, serial_str.as_ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(DiscoveryMessage { ptr: ret })
        }
    }

    /// Connect to the camera and gain exclusive access
    ///
    /// ### Arguments
    ///
    /// - `message` Discovery message of the camera to connect to
    ///
    pub fn connect(&mut self, message: &DiscoveryMessage) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_connect(self.ptr, message.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Disconnect from the camera
    ///
    pub fn disconnect(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_disconnect(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Check whether a connection is still maintained with the camera
    ///
    /// ### Returns
    ///
    /// Is connected
    ///
    pub fn is_connected(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_is_connected(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get stream packet size
    ///
    /// ### Returns
    ///
    /// The packet size
    ///
    pub fn get_stream_packet_size(&self) -> Result<u16> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_stream_packet_size(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set stream packet size
    ///
    /// ### Arguments
    ///
    /// - `packet_size` The packet size
    ///
    pub fn set_stream_packet_size(&mut self, packet_size: u16) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_set_stream_packet_size(self.ptr, packet_size, err.ptr()) };
        err.empty_result()
    }

    /// Test whether using the set packet size works
    ///
    /// ### Arguments
    ///
    /// - `port` Port to send test packets to
    ///
    /// ### Returns
    ///
    /// Packet size works
    ///
    pub fn test_stream_packet_size(&mut self, opt_port: Option<u16>) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        let ret = unsafe { tof_gige_interface_test_stream_packet_size(self.ptr, port, err.ptr()) };
        err.make_result(ret)
    }

    /// Get delay between network packets
    ///
    /// ### Returns
    ///
    /// The delay, depending on the camera software it is in nanoseconds or in
    /// cpu instruction
    ///
    pub fn get_delay(&self) -> Result<u32> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_delay(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set delay between network packets
    ///
    /// ### Arguments
    ///
    /// - `delay` The delay, depending on the camera software it is in
    ///
    pub fn set_delay(&mut self, delay: u32) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_set_delay(self.ptr, delay, err.ptr()) };
        err.empty_result()
    }

    /// Download camera configuration from the camera
    ///
    /// ### Arguments
    ///
    /// - `port` Port to download configuration on
    ///
    /// ### Returns
    ///
    /// Camera configuration
    ///
    pub fn download_configuration(&mut self, opt_port: Option<u16>) -> Result<CameraConfig> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        let ret = unsafe { tof_gige_interface_download_configuration(self.ptr, port, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

    /// Download calibration from the camera
    ///
    /// ### Arguments
    ///
    /// - `port` Port to download configuration on
    ///
    /// ### Returns
    ///
    /// Calibration
    ///
    pub fn download_calibration(&mut self, opt_port: Option<u16>) -> Result<Calibration> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        let ret = unsafe { tof_gige_interface_download_calibration(self.ptr, port, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Calibration { ptr: ret })
        }
    }

    /// Upload camera configuration to the camera
    ///
    /// ### Arguments
    ///
    /// - `config` Camera configuration
    /// - `port` Port to download configuration on
    ///
    pub fn upload_configuration(&mut self, config: &CameraConfig, opt_port: Option<u16>) -> Result<()> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        unsafe { tof_gige_interface_upload_configuration(self.ptr, config.ptr(), port, err.ptr()) };
        err.empty_result()
    }

    /// Check whether this camera is capable of streaming the specified type
    ///
    /// ### Arguments
    ///
    /// - `stream` Depth stream type
    ///
    /// ### Returns
    ///
    /// Is capable
    ///
    pub fn depth_stream_capable(&self, stream: DepthStreamType) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_depth_stream_capable(self.ptr, stream as i32, err.ptr()) };
        err.make_result(ret)
    }

    /// Start the depth stream
    ///
    /// ### Arguments
    ///
    /// - `stream` Depth stream type
    /// - `port` Port to receive depth data on
    ///
    pub fn start_depth_stream(&mut self, stream: DepthStreamType, opt_port: Option<u16>) -> Result<()> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        unsafe { tof_gige_interface_start_depth_stream(self.ptr, stream as i32, port, err.ptr()) };
        err.empty_result()
    }

    /// Stop the depth stream
    ///
    pub fn stop_depth_stream(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_stop_depth_stream(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Returns true if a depth frame is available
    ///
    /// ### Returns
    ///
    /// Depth frame available
    ///
    pub fn has_depth_data(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_has_depth_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get a depth frame
    ///
    /// ### Returns
    ///
    /// Depth frame
    ///
    pub fn get_depth_data(&mut self) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_depth_data(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Check whether camera supports an image stream
    ///
    /// ### Returns
    ///
    /// Camera supports an image stream
    ///
    pub fn image_stream_capable(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_image_stream_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Start image stream
    ///
    /// ### Arguments
    ///
    /// - `port` Port to receive image data on
    ///
    pub fn start_image_stream(&mut self, opt_port: Option<u16>) -> Result<()> {
        let mut err = ErrorHelper::new();
        let port = opt_port.unwrap_or(0);
        unsafe { tof_gige_interface_start_image_stream(self.ptr, port, err.ptr()) };
        err.empty_result()
    }

    /// Stop image stream
    ///
    pub fn stop_image_stream(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_stop_image_stream(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Check whether image frame is available
    ///
    /// ### Returns
    ///
    /// Image frame available
    ///
    pub fn has_image_data(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_has_image_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get image frame
    ///
    /// ### Returns
    ///
    /// Image data
    ///
    pub fn get_image_data(&mut self) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_image_data(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get the image frame into the supplied pointer
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    ///
    /// ### Returns
    ///
    /// Image data
    ///
    pub unsafe fn get_image_data_into_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = tof_gige_interface_get_image_data_into_pointer(self.ptr, pointer, capacity, callback, callback_user_ptr, err.ptr());
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get whether the camera is persistent ip capable
    ///
    /// ### Returns
    ///
    /// Persistent IP capable
    ///
    pub fn persistent_ip_capable(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_persistent_ip_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get the current persistent ip settings
    ///
    /// ### Returns
    ///
    /// Persistent IP
    ///
    pub fn get_persistent_ip(&self) -> Result<PersistentIp> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_persistent_ip(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(PersistentIp { ptr: ret })
        }
    }

    /// Set the persistent ip settings. The persistent ip settings will be
    /// active on next reboot.
    ///
    /// ### Arguments
    ///
    /// - `persistent_ip` The persistent ip
    ///
    pub fn set_persistent_ip(&mut self, persistent_ip: &PersistentIp) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_set_persistent_ip(self.ptr, persistent_ip.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Get the capacity for amount of user pointers that can be registered
    ///
    /// ### Returns
    ///
    /// User pointer capacity
    ///
    pub fn get_user_pointer_capacity(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_get_user_pointer_capacity(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the capacity for amount of user pointers that can be registered
    ///
    /// ### Arguments
    ///
    /// - `capacity` User pointer capacity
    ///
    pub fn set_user_pointer_capacity(&mut self, capacity: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_set_user_pointer_capacity(self.ptr, capacity, err.ptr()) };
        err.empty_result()
    }

    /// Add a pointer that will be filled with depth data. This method allows
    /// you to fill data straight into your own data structure without any
    /// additional copying.
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    ///
    pub unsafe fn add_depth_user_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void) -> Result<()> {
        let mut err = ErrorHelper::new();
        tof_gige_interface_add_depth_user_pointer(self.ptr, pointer, capacity, callback, callback_user_ptr, err.ptr());
        err.empty_result()
    }

    /// Get the tof library version running on the camera
    ///
    /// ### Returns
    ///
    /// Version
    ///
    pub fn version(&mut self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_version(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Check whether the camera is capable of software trigger
    ///
    /// ### Returns
    ///
    /// Software trigger capable
    ///
    pub fn software_trigger_capable(&mut self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_gige_interface_software_trigger_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Software trigger the camera
    ///
    pub fn software_trigger(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_gige_interface_software_trigger(self.ptr, err.ptr()) };
        err.empty_result()
    }

}

#[repr(C)]
struct tof_usb_device {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_usb_device_delete(_: *mut tof_usb_device);
    fn tof_usb_device_serial(_: *mut tof_usb_device, _: *mut *mut tof_error) -> *const c_char;
}

/// USB device
pub struct UsbDevice {
    ptr: *mut tof_usb_device,
}

impl Drop for UsbDevice {
    fn drop(&mut self) {
        unsafe { tof_usb_device_delete(self.ptr); }
    }
}

unsafe impl Send for UsbDevice {}

impl UsbDevice {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_usb_device {
        self.ptr
    }

    /// Get serial corresponding with USB device
    ///
    /// ### Returns
    ///
    /// Serial
    ///
    pub fn serial(&mut self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_device_serial(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

}

#[repr(C)]
struct tof_usb_interface {
    _data: [u8; 0],
    _marker: core::marker::PhantomData<(*mut u8, core::marker::PhantomPinned)>,
}

#[link(name = "tof")]
extern "C" {
    fn tof_usb_interface_delete(_: *mut tof_usb_interface);
    fn tof_usb_interface_new(_: *mut tof_usb_device, _: *mut *mut tof_error) -> *mut tof_usb_interface;
    fn tof_usb_interface_is_connected(_: *const tof_usb_interface, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_test_stream_speed(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> f64;
    fn tof_usb_interface_download_configuration(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> *mut tof_camera_config;
    fn tof_usb_interface_download_calibration(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> *mut tof_calibration;
    fn tof_usb_interface_upload_configuration(_: *mut tof_usb_interface, _: *const tof_camera_config, _: *mut *mut tof_error);
    fn tof_usb_interface_depth_stream_capable(_: *const tof_usb_interface, _: i32, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_start_depth_stream(_: *mut tof_usb_interface, _: i32, _: *mut *mut tof_error);
    fn tof_usb_interface_stop_depth_stream(_: *mut tof_usb_interface, _: *mut *mut tof_error);
    fn tof_usb_interface_has_depth_data(_: *const tof_usb_interface, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_get_depth_data(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_usb_interface_image_stream_capable(_: *const tof_usb_interface, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_start_image_stream(_: *mut tof_usb_interface, _: *mut *mut tof_error);
    fn tof_usb_interface_stop_image_stream(_: *mut tof_usb_interface, _: *mut *mut tof_error);
    fn tof_usb_interface_has_image_data(_: *const tof_usb_interface, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_get_image_data(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_usb_interface_get_image_data_into_pointer(_: *mut tof_usb_interface, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: *mut *mut tof_error) -> *mut tof_data;
    fn tof_usb_interface_get_user_pointer_capacity(_: *const tof_usb_interface, _: *mut *mut tof_error) -> usize;
    fn tof_usb_interface_set_user_pointer_capacity(_: *mut tof_usb_interface, _: usize, _: *mut *mut tof_error);
    fn tof_usb_interface_add_depth_user_pointer(_: *mut tof_usb_interface, _: *mut u8, _: usize, _: UserPointerDestructed, _: *mut c_void, _: *mut *mut tof_error);
    fn tof_usb_interface_version(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> *const c_char;
    fn tof_usb_interface_software_trigger_capable(_: *mut tof_usb_interface, _: *mut *mut tof_error) -> bool;
    fn tof_usb_interface_software_trigger(_: *mut tof_usb_interface, _: *mut *mut tof_error);
}

/// DEPRECATED, use KeaCamera. Connect to a camera directly using the USB
/// interface
pub struct UsbInterface {
    ptr: *mut tof_usb_interface,
}

impl Drop for UsbInterface {
    fn drop(&mut self) {
        unsafe { tof_usb_interface_delete(self.ptr); }
    }
}

unsafe impl Send for UsbInterface {}

impl UsbInterface {
    #[allow(dead_code)]
    fn ptr(&self) -> *mut tof_usb_interface {
        self.ptr
    }

    /// Create USB camera from a usb device
    ///
    /// ### Arguments
    ///
    /// - `usb_device` USB device
    ///
    pub fn new(usb_device: &UsbDevice) -> Result<Self> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_new(usb_device.ptr(), err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(UsbInterface { ptr: ret })
        }
    }

    /// Check whether a connection is still maintained with the camera
    ///
    /// ### Returns
    ///
    /// Is connected
    ///
    pub fn is_connected(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_is_connected(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Tests the connection speed of the USB port. This is useful to check
    /// whether a hub or USB port works well. The maximum speed that the camera
    /// is able to transmit is ~3.0 Gbps.
    ///
    /// ### Returns
    ///
    /// Connection speed in Gpbs (bits)
    ///
    pub fn test_stream_speed(&mut self) -> Result<f64> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_test_stream_speed(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Download camera configuration from the camera
    ///
    /// ### Returns
    ///
    /// Camera configuration
    ///
    pub fn download_configuration(&mut self) -> Result<CameraConfig> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_download_configuration(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(CameraConfig { ptr: ret })
        }
    }

    /// Download calibration from the camera
    ///
    /// ### Returns
    ///
    /// Calibration
    ///
    pub fn download_calibration(&mut self) -> Result<Calibration> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_download_calibration(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Calibration { ptr: ret })
        }
    }

    /// Upload camera configuration to the camera
    ///
    /// ### Arguments
    ///
    /// - `config` Camera configuration
    ///
    pub fn upload_configuration(&mut self, config: &CameraConfig) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_upload_configuration(self.ptr, config.ptr(), err.ptr()) };
        err.empty_result()
    }

    /// Check whether this camera is capable of streaming the specified type
    ///
    /// ### Arguments
    ///
    /// - `stream` Depth stream type
    ///
    /// ### Returns
    ///
    /// Is capable
    ///
    pub fn depth_stream_capable(&self, stream: DepthStreamType) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_depth_stream_capable(self.ptr, stream as i32, err.ptr()) };
        err.make_result(ret)
    }

    /// Start the depth stream
    ///
    /// ### Arguments
    ///
    /// - `stream` Depth stream type
    ///
    pub fn start_depth_stream(&mut self, stream: DepthStreamType) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_start_depth_stream(self.ptr, stream as i32, err.ptr()) };
        err.empty_result()
    }

    /// Stop the depth stream
    ///
    pub fn stop_depth_stream(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_stop_depth_stream(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Returns true if a depth frame is available
    ///
    /// ### Returns
    ///
    /// Depth frame available
    ///
    pub fn has_depth_data(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_has_depth_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get a depth frame
    ///
    /// ### Returns
    ///
    /// Depth frame
    ///
    pub fn get_depth_data(&mut self) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_get_depth_data(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Check whether camera supports an image stream
    ///
    /// ### Returns
    ///
    /// Camera supports an image stream
    ///
    pub fn image_stream_capable(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_image_stream_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Start image stream
    ///
    pub fn start_image_stream(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_start_image_stream(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Stop image stream
    ///
    pub fn stop_image_stream(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_stop_image_stream(self.ptr, err.ptr()) };
        err.empty_result()
    }

    /// Check whether image frame is available
    ///
    /// ### Returns
    ///
    /// Image frame available
    ///
    pub fn has_image_data(&self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_has_image_data(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Get image frame
    ///
    /// ### Returns
    ///
    /// Image data
    ///
    pub fn get_image_data(&mut self) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_get_image_data(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get the image frame into the supplied pointer
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    ///
    /// ### Returns
    ///
    /// Image data
    ///
    pub unsafe fn get_image_data_into_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void) -> Result<Data> {
        let mut err = ErrorHelper::new();
        let ret = tof_usb_interface_get_image_data_into_pointer(self.ptr, pointer, capacity, callback, callback_user_ptr, err.ptr());
        if err.has_ptr() {
            Err(err.make_error())
        } else {
            Ok(Data { ptr: ret })
        }
    }

    /// Get the capacity for amount of user pointers that can be registered
    ///
    /// ### Returns
    ///
    /// User pointer capacity
    ///
    pub fn get_user_pointer_capacity(&self) -> Result<usize> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_get_user_pointer_capacity(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Set the capacity for amount of user pointers that can be registered
    ///
    /// ### Arguments
    ///
    /// - `capacity` User pointer capacity
    ///
    pub fn set_user_pointer_capacity(&mut self, capacity: usize) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_set_user_pointer_capacity(self.ptr, capacity, err.ptr()) };
        err.empty_result()
    }

    /// Add a pointer that will be filled with depth data. This method allows
    /// you to fill data straight into your own data structure without any
    /// additional copying.
    ///
    /// ### Arguments
    ///
    /// - `pointer` The pointer to the data to write into
    /// - `capacity` The amount of data the pointer can hold
    /// - `callback` Callback that will be called when pointer is no longer in
    ///
    pub unsafe fn add_depth_user_pointer(&mut self, pointer: *mut u8, capacity: usize, callback: UserPointerDestructed, callback_user_ptr: *mut c_void) -> Result<()> {
        let mut err = ErrorHelper::new();
        tof_usb_interface_add_depth_user_pointer(self.ptr, pointer, capacity, callback, callback_user_ptr, err.ptr());
        err.empty_result()
    }

    /// Get the tof library version running on the camera
    ///
    /// ### Returns
    ///
    /// Version
    ///
    pub fn version(&mut self) -> Result<&str> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_version(self.ptr, err.ptr()) };
        if err.has_ptr() {
            Err(err.make_error())
        } else {
        Ok(unsafe { CStr::from_ptr(ret).to_str()? })
        }
    }

    /// Check whether the camera is capable of software trigger
    ///
    /// ### Returns
    ///
    /// Software trigger capable
    ///
    pub fn software_trigger_capable(&mut self) -> Result<bool> {
        let mut err = ErrorHelper::new();
        let ret = unsafe { tof_usb_interface_software_trigger_capable(self.ptr, err.ptr()) };
        err.make_result(ret)
    }

    /// Software trigger the camera
    ///
    pub fn software_trigger(&mut self) -> Result<()> {
        let mut err = ErrorHelper::new();
        unsafe { tof_usb_interface_software_trigger(self.ptr, err.ptr()) };
        err.empty_result()
    }

}

#[link(name = "tof")]
extern "C" {
    fn tof_usb_device_discover(_: *const *mut tof_usb_device, _: usize, _: *mut *mut tof_error) -> usize;
    fn tof_usb_device_discover_one(_: *mut *mut tof_error) -> *mut tof_usb_device;
    fn tof_usb_device_find(_: *const c_char, _: *mut *mut tof_error) -> *mut tof_usb_device;
}

/// Discover all connected chronoptics usb cameras
///
/// ### Returns
///
/// The usb devices
///
pub fn usb_device_discover() -> Result<Vec<UsbDevice>> {
    let mut err = ErrorHelper::new();
    let size = unsafe { tof_usb_device_discover(std::ptr::null_mut(), 0, err.ptr()) };
    if err.has_ptr() {
        return Err(err.make_error());
    }
    let mut out: Vec<UsbDevice> = Vec::with_capacity(size);
    let ret = unsafe { tof_usb_device_discover(std::ptr::addr_of!((*out.as_mut_ptr()).ptr), out.capacity(), err.ptr()) };
    unsafe { out.set_len(ret) };
    err.make_result(out)
}

/// Discover one chronoptics usb camera
///
/// ### Returns
///
/// A USB device
///
pub fn usb_device_discover_one() -> Result<UsbDevice> {
    let mut err = ErrorHelper::new();
    let ret = unsafe { tof_usb_device_discover_one(err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(UsbDevice { ptr: ret })
    }
}

/// Discover usb camera with the given serial
///
/// ### Arguments
///
/// - `serial` Serial of the camera to find
///
/// ### Returns
///
/// A USB device
///
pub fn usb_device_find(serial: &str) -> Result<UsbDevice> {
    let mut err = ErrorHelper::new();
    let serial_str = CString::new(serial)?;
    let ret = unsafe { tof_usb_device_find(serial_str.as_ptr(), err.ptr()) };
    if err.has_ptr() {
        Err(err.make_error())
    } else {
        Ok(UsbDevice { ptr: ret })
    }
}

pub trait DataType: Copy + private::Private {
    fn has_frame_type(frame_type: FrameType) -> bool;
}

macro_rules! data_type {
    ($rust_type: ty, $frame_types: expr) => {
        impl $crate::DataType for $rust_type {
            fn has_frame_type(frame_type: FrameType) -> bool {
                let frame_types = $frame_types;
                frame_types.contains(&frame_type)
            }
        }

        impl private::Private for $rust_type {}
    };
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct XYZ {
  pub x: f32,
  pub y: f32,
  pub z: f32,
  padding: f32
}

impl XYZ {
  pub fn new(x: f32, y: f32, z: f32) -> XYZ {
    XYZ{x, y, z, padding: 0.0}
  } 
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct XYZAmp {
  pub x: f32,
  pub y: f32,
  pub z: f32,
  pub amplitude: f32
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct XYZBGR {
  pub x: f32,
  pub y: f32,
  pub z: f32,
  pub b: u8,
  pub g: u8,
  pub r: u8,
  padding: u8,
}

impl XYZBGR {
  pub fn new(x: f32, y: f32, z: f32, b: u8, g: u8, r: u8) -> XYZBGR {
    XYZBGR{x, y, z, b, g, r, padding: 0} 
  }
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct XYZBGRI {
  pub x: f32,
  pub y: f32,
  pub z: f32,
  pub b: u8,
  pub g: u8,
  pub r: u8,
  pub intensity: u8,
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct BGR {
  pub b: u8,
  pub g: u8,
  pub r: u8,
}

data_type!(u8, [FrameType::Intensity, FrameType::Yuv, FrameType::Mjpeg]);
data_type!(i16, [FrameType::Raw]);
data_type!(u16, [FrameType::RawAb, FrameType::RawCommon, FrameType::Radial]);
data_type!(f32, [FrameType::Real, FrameType::Imaginary, FrameType::Phase, FrameType::Amplitude, FrameType::X, FrameType::Y, FrameType::Z, ]);
data_type!(XYZ, [FrameType::Xyz]);
data_type!(XYZAmp, [FrameType::XyzAmp, FrameType::CompressedDot]);
data_type!(XYZBGR, [FrameType::XyzBgr]);
data_type!(XYZBGRI, [FrameType::XyzBgrI]);
data_type!(BGR, [FrameType::Bgr, FrameType::BgrProjected]);

impl Data {
    pub fn len<T: DataType>(&mut self) -> Result<usize> {
      let frame_type = self.frame_type()?;
      if !T::has_frame_type(frame_type) {
        Err(Error::new("Frame types don't correspond"))
      } else {
        Ok(self.size()? / std::mem::size_of::<T>())
      }
    }

    pub fn as_slice<T: DataType>(&self) -> Result<&[T]> {
        let frame_type = self.frame_type()?;
        if !T::has_frame_type(frame_type) {
            Err(Error::new("Frame types don't correspond"))
        } else {
            unsafe { self.as_slice_unchecked() }
        }
    }

    /// # Safety
    /// Caller must ensure that the `T` type argument corresponds to the data stored in `Data`
    pub unsafe fn as_slice_unchecked<T>(&self) -> Result<&[T]> {
        let data = self.data()?;
        if data.is_null() {
            Err(Error::new("No data is contained"))
        } else {
            let len = self.size()? / std::mem::size_of::<T>();
            let out = std::ptr::slice_from_raw_parts::<T>(data as *const T, len);
            Ok(&*out)
        }
    }

    pub fn as_mut_slice<T: DataType>(&mut self) -> Result<&mut [T]> {
        let frame_type = self.frame_type()?;
        if !T::has_frame_type(frame_type) {
            Err(Error::new("Frame types don't correspond"))
        } else {
            unsafe { self.as_mut_slice_unchecked() }
        }
    }

    /// # Safety
    /// Caller must ensure that the `T` type argument corresponds to the data stored in `Data`
    pub unsafe fn as_mut_slice_unchecked<T>(&mut self) -> Result<&mut [T]> {
        let data = self.data()?;
        if data.is_null() {
            Err(Error::new("No data is contained"))
        } else {
            let len = self.size()? / std::mem::size_of::<T>();
            let out = std::ptr::slice_from_raw_parts_mut::<T>(data as *mut T, len);
            Ok(&mut *out)
        }
    }
}

